"""
Copyright 2023 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import abc
import logging
import logging.config
import os
import re
import sys
from argparse import Namespace
from collections.abc import Mapping, Sequence, Set
from io import TextIOWrapper
from logging import handlers
from typing import Optional, TextIO

import colorlog
import yaml
from colorlog.formatter import LogColors
from yaml import Dumper, Node

from inmanta import config, const
from inmanta.config import Option, component_log_configs, logging_config
from inmanta.const import LOG_CONTEXT_VAR_ENVIRONMENT, NAME_RESOURCE_ACTION_LOGGER
from inmanta.server import config as server_config
from inmanta.stable_api import stable_api

logfire_enabled = os.getenv("LOGFIRE_TOKEN", None) is not None
try:
    from logfire.integrations.logging import LogfireLoggingHandler
except ModuleNotFoundError:
    logfire_enabled = False

LOGGER = logging.getLogger(__name__)


def _is_on_tty() -> bool:
    return (hasattr(sys.stdout, "isatty") and sys.stdout.isatty()) or const.ENVIRON_FORCE_TTY in os.environ


def python_log_level_to_name(python_log_level: int) -> str:
    """Convert a python log level to a human readable version that works in log config files"""
    # we build the reverse mapping every time because
    # we don't want to use private fields
    # name_to_level = logging._levelToName
    # the underlying assumption is that this code is not performance critical
    name_to_level = logging.getLevelNamesMapping()
    level_to_name = {v: k for k, v in name_to_level.items()}

    result = level_to_name.get(python_log_level)
    if result is not None:
        return result
    return str(python_log_level)


def python_log_level_to_int(level: int | str) -> int:
    # From python logging framework _checkLevel, copied to not use their private methods

    name_to_level = logging.getLevelNamesMapping()
    if isinstance(level, int):
        rv = level
    elif str(level) == level:
        if level not in name_to_level:
            raise ValueError("Unknown level: %r" % level)
        rv = name_to_level[level]
    else:
        raise TypeError("Level not an integer or a valid string: %r" % (level,))
    return rv


"""
This dictionary maps the Inmanta log levels to the corresponding Python log levels
"""
log_levels = {
    "0": logging.ERROR,
    "1": logging.WARNING,
    "2": logging.INFO,
    "3": logging.DEBUG,
    "4": 3,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "TRACE": 3,
}

logging.addLevelName(3, "TRACE")


class NoLoggingConfigFound(Exception):
    pass


class LogConfigDumper(Dumper):
    """
    The representer config is class level

    If we don't subclass, we re-configure the every yaml serializer for the entire process

    To prevent this, we subclass"""

    def encode_streams(self, data: object) -> Node:
        if data == sys.stdout:
            return self.represent_data("ext://sys.stdout")
        if data == sys.stderr:
            return self.represent_data("ext://sys.stderr")
        raise Exception(f"Can not encode stream {data}")


LogConfigDumper.add_representer(TextIOWrapper, LogConfigDumper.encode_streams)


@stable_api
class LoggingConfigExtension:
    """
    This class is used by an extension to declare the default logging config that should be used when
    no dict-based logging config was provided by the user. The FullLoggingConfig class provides support
    to merge this logging config into the main logging config of inmanta-core. This class supports only
    version 1 of Python's dictConfig format.
    """

    def __init__(
        self,
        *,
        formatters: Optional[Mapping[str, object]] = None,
        handlers: Optional[Mapping[str, object]] = None,
        loggers: Optional[Mapping[str, object]] = None,
        root_handlers: Optional[list[str]] = None,
        log_dirs_to_create: Optional[Set[str]] = None,
    ) -> None:
        """
        :param log_dirs_to_create: The log directories that should be created before the logging config can be used.
        """
        self.formatters = formatters if formatters else {}
        self.handlers = handlers if handlers else {}
        self.loggers = loggers if loggers else {}
        self.root_handlers = root_handlers if root_handlers else []
        self.log_dirs_to_create = log_dirs_to_create if log_dirs_to_create else set()

    def ensure_log_dirs(self) -> None:
        """
        This method makes sure that the log directories, required by this logging config, are present on disk.
        The directories are created if they don't exist yet.
        """
        for directory in self.log_dirs_to_create:
            os.makedirs(directory, exist_ok=True)


class FullLoggingConfig(LoggingConfigExtension):
    """
    A FullLoggingConfig that can be applied on Python's logging framework.

    This class supports only version 1 of Python's dictConfig format.
    """

    def __init__(
        self,
        *,
        formatters: Optional[Mapping[str, object]] = None,
        handlers: Optional[Mapping[str, object]] = None,
        loggers: Optional[Mapping[str, object]] = None,
        root_handlers: Optional[list[str]] = None,
        log_dirs_to_create: Optional[Set[str]] = None,
        root_log_level: Optional[int | str] = None,
    ):
        super().__init__(
            formatters=formatters,
            handlers=handlers,
            loggers=loggers,
            root_handlers=root_handlers,
            log_dirs_to_create=log_dirs_to_create,
        )
        self.root_log_level = root_log_level

    def join(self, logging_config_extension: LoggingConfigExtension, allow_overwrite: bool = False) -> "FullLoggingConfig":
        """
        Join this FullLoggingConfig and the given LoggingConfigExtension together into a single
        FullLoggingConfig object that contains all the logging config of both objects.
        """

        def warn_or_raise_on_common(component: str, one_set: Mapping[str, object], other_set: Mapping[str, object]) -> None:
            common_names = set(one_set.keys()) & set(other_set.keys())
            # don't warn if identical!
            common_names = {
                common_name for common_name in common_names if one_set.get(common_name) != other_set.get(common_name)
            }
            if not common_names:
                return

            if not allow_overwrite:
                raise Exception(f"The following {component} names appear in multiple logging configs: {common_names}")
            else:
                logging.warning("The following %s names appear in multiple logging configs: %s", component, common_names)

        warn_or_raise_on_common("formatter", self.formatters, logging_config_extension.formatters)
        warn_or_raise_on_common("handler", self.handlers, logging_config_extension.handlers)
        warn_or_raise_on_common("logger", self.loggers, logging_config_extension.loggers)

        def update_join(one_dict: Mapping[str, object], two_dict: Mapping[str, object]) -> dict[str, object]:
            out = dict(**one_dict)
            out.update(two_dict)
            return out

        root_level = self.root_log_level
        if isinstance(logging_config_extension, FullLoggingConfig):
            if logging_config_extension.root_log_level is not None:
                if root_level is None:
                    # base config has no root level?
                    root_level = logging_config_extension.root_log_level
                else:
                    # take lowest
                    root_level_int_one = python_log_level_to_int(root_level)
                    root_level_int_other = python_log_level_to_int(logging_config_extension.root_log_level)
                    root_level = python_log_level_to_name(min(root_level_int_one, root_level_int_other))

        return FullLoggingConfig(
            formatters=update_join(self.formatters, logging_config_extension.formatters),
            handlers=update_join(self.handlers, logging_config_extension.handlers),
            loggers=update_join(self.loggers, logging_config_extension.loggers),
            root_handlers=sorted(list(set(self.root_handlers) | set(logging_config_extension.root_handlers))),
            log_dirs_to_create=self.log_dirs_to_create | logging_config_extension.log_dirs_to_create,
            root_log_level=root_level,
        )

    def apply_config(self) -> None:
        """
        Configure the logging system with this logging config.
        """
        self.ensure_log_dirs()
        dict_config = self._to_dict_config()
        logging.config.dictConfig(dict_config)

    def _to_dict_config(self) -> dict[str, object]:
        """
        Convert this object into a dictionary format that can be passed to logging.config.dictConfig() method
        to configure logging.
        """
        return {
            "version": 1,
            "formatters": dict(self.formatters),
            "handlers": dict(self.handlers),
            "loggers": dict(self.loggers),
            "root": {
                "handlers": self.root_handlers,
                **({"level": self.root_log_level} if self.root_log_level else {}),
            },
            "disable_existing_loggers": False,
        }

    def to_string(self) -> str:
        return yaml.dump(self._to_dict_config(), Dumper=LogConfigDumper)


class Options(Namespace):
    """
    The Options class provides a way to configure the InmantaLoggerConfig with the following attributes:

    :param log_file: if this attribute is set, the logs will be written to the specified file instead of the stream
                     specified in `get_instance`.
    :param log_file_level: the Inmanta logging level for the file handler (if `log_file` is set).
                           The possible inmanta log levels and their associated python log level are defined in the
                           inmanta.logging.log_levels dictionary.
    :param verbose: the verbosity level of the log messages. can be a number from 0 to 4.
                    if a bigger number is provided, 4 will be used. Refer to log_file_level for the explanation of each level.
                    default is 1 (WARNING)
    :param timed: if true,  adds the time to the formatter in the log lines.
    :param keep_logger_names: Display the log messages using the name of the logger that created the log message,
                              instead of the component of the compiler that was executing while the log record was created
                              or the name of the module that created the log message.
    :param logging_config: Path to the dict-based logging config file.
    """

    log_file: Optional[str] = None
    log_file_level: str = "INFO"
    verbose: int = 1
    timed: bool = False
    keep_logger_names: bool = False
    logging_config: Optional[str] = None


class LoggingConfigBuilderExtension(abc.ABC):

    @abc.abstractmethod
    def get_logging_config_from_options(
        self,
        stream: TextIO,
        options: Options,
        component: str | None,
        context: Mapping[str, str],
        master_config: FullLoggingConfig,
    ) -> FullLoggingConfig:
        """
        Update the existing config with additional configuration for this extension
        """
        pass


class LoggingConfigBuilder:

    def get_bootstrap_logging_config(
        self,
        stream: TextIO = sys.stdout,
        python_log_level: int = logging.INFO,
    ) -> FullLoggingConfig:
        """
        This method returns the logging config that should be used between the moment that the process starts,
        and the moment that the logging-related config options are parsed and applied.

        :param stream: The TextIO stream where the logs will be sent to.
        :param python_log_level: python log level to configure for the bootstrap logger
        """
        name_root_handler = "core_console_handler"
        log_level_name = python_log_level_to_name(python_log_level)
        logging_config_core = FullLoggingConfig(
            formatters={
                "core_console_formatter": self._get_multiline_formatter_config(True),
            },
            handlers={
                name_root_handler: {
                    "class": "logging.StreamHandler",
                    "formatter": "core_console_formatter",
                    "level": log_level_name,
                    "stream": stream,
                },
            },
            loggers={},
            root_handlers=[name_root_handler],
            root_log_level=log_level_name,
        )
        return logging_config_core

    @classmethod
    def get_log_file_for_scheduler(cls, env: str, log_dir: str, log_file_cli_option: str | None = None) -> str:
        """
        Returns the path to the main log file of the scheduler.

        :param env: The environment parameter to use in the name of the log file.
        :param log_dir: The log directory as configured using the config.log_dir configuration option.
        :param log_file_cli_option: The value of the --log-file CLI config option or None if the option was not provided.
        """
        if log_file_cli_option:
            return log_file_cli_option
        else:
            # use setting as formerly passed by the autostarted agent manager if not set via CLI
            return os.path.join(log_dir, f"agent-{env}.log")

    def get_logging_config_from_options(
        self,
        stream: TextIO,
        options: Options,
        component: str | None,
        context: Mapping[str, str],
    ) -> FullLoggingConfig:
        """
        Return the logging config based on the given configuration options, passed on the CLI,
        and the configuration options present in the config file.

        :param stream: The TextIO stream where the logs will be sent to.
        :param options: The config options passed on the CLI.
        :param component: component we are starting
        :param context: the component context we are starting with
        """
        handlers: dict[str, object] = {}
        loggers: dict[str, object] = {}
        handler_root_logger: str
        log_level: int

        log_file: Optional[str] = options.log_file

        short_names = False
        if component == "compiler":
            short_names = not options.keep_logger_names

        elif component == "scheduler":
            # Override defaults
            if LOG_CONTEXT_VAR_ENVIRONMENT not in context:
                raise Exception("The scheduler expects an environment as context")

            env = context[LOG_CONTEXT_VAR_ENVIRONMENT]

            log_file = self.get_log_file_for_scheduler(
                env=env, log_dir=config.log_dir.get(), log_file_cli_option=options.log_file
            )

            # We don't override log-file-level as we can't detect if it is set

        # Shared config
        if log_file:
            log_level = convert_inmanta_log_level(options.log_file_level)
            handler_root_logger = f"{component}_handler" if component is not None else "root_handler"
            handlers[handler_root_logger] = {
                "class": "logging.handlers.WatchedFileHandler",
                "level": python_log_level_to_name(log_level),
                "formatter": "core_log_formatter",
                "filename": log_file,
                "mode": "a+",
            }
        else:
            log_level = convert_inmanta_log_level(inmanta_log_level=str(options.verbose), cli=True)
            handler_root_logger = "core_console_handler"
            handlers[handler_root_logger] = {
                "class": "logging.StreamHandler",
                "formatter": "core_console_formatter",
                "level": python_log_level_to_name(log_level),
                "stream": stream,
            }

        formatters = {
            # Always add all the formatters, even if they are not used by configuration. This way
            # the formatters can be used if the user dumps the default logging config to file.
            "core_console_formatter": self._get_multiline_formatter_config(not short_names, options),
            "core_log_formatter": {
                "format": "%(asctime)s %(levelname)-8s %(name)-10s %(message)s",
            },
        }

        handlers.update(
            {
                "core_tornado_debug_log_handler": {
                    "class": "inmanta.logging.TornadoDebugLogHandler",
                    "level": "DEBUG",
                },
            }
        )

        loggers.update(
            {
                "tornado.general": {
                    "level": "DEBUG",
                    "propagate": True,
                    "handlers": ["core_tornado_debug_log_handler"],
                }
            }
        )

        if component == "server":
            # Fully generic
            pass
        elif component == "scheduler":
            # Resource action log
            handlers.update(
                {
                    "scheduler_resource_action_handler": {
                        "class": "logging.handlers.WatchedFileHandler",
                        "level": "DEBUG",
                        "formatter": "core_log_formatter",
                        "filename": os.path.join(
                            config.log_dir.get(),
                            server_config.server_resource_action_log_prefix.get()
                            + f"{context.get(LOG_CONTEXT_VAR_ENVIRONMENT)}.log",
                        ),
                    },
                }
            )
            loggers.update(
                {
                    NAME_RESOURCE_ACTION_LOGGER: {
                        "level": "DEBUG",
                        "propagate": True,
                        "handlers": ["scheduler_resource_action_handler"],
                    },
                }
            )

        full_logging_config = FullLoggingConfig(
            formatters=formatters,
            handlers=handlers,
            loggers=loggers,
            root_handlers=[handler_root_logger],
            root_log_level=python_log_level_to_name(log_level),
        )

        return full_logging_config

    def _get_multiline_formatter_config(self, keep_logger_names: bool, options: Optional[Options] = None) -> dict[str, object]:
        """
        Returns the dict-based formatter config for logs that will be sent to the console.

        :param options: The config options requested by the user or None if the config options are not parsed yet and
                        the bootstrap_logger_config should be used.
        """

        # Use a shorter space padding if we know that we will use short names as the logger name.
        # Otherwise the log records contains too much white spaces.
        space_padding_after_logger_name = 25 if (keep_logger_names) else 15
        log_format = "%(asctime)s " if options and options.timed else ""
        if _is_on_tty():
            log_format += f"%(log_color)s%(name)-{space_padding_after_logger_name}s%(levelname)-8s%(reset)s%(blue)s%(message)s"
            log_colors = {"DEBUG": "cyan", "INFO": "green", "WARNING": "yellow", "ERROR": "red", "CRITICAL": "red"}
        else:
            log_format += f"%(name)-{space_padding_after_logger_name}s%(levelname)-8s%(message)s"
            log_colors = None

        return {
            "()": "inmanta.logging.MultiLineFormatter",
            "fmt": log_format,
            "log_colors": log_colors,
            "reset": _is_on_tty(),
            "no_color": not _is_on_tty(),
            "keep_logger_names": keep_logger_names,
        }


@stable_api
def convert_inmanta_log_level(inmanta_log_level: str, cli: bool = False) -> int:
    """
    Convert the given Inmanta log level to the corresponding Python log level.

    :param inmanta_log_level: The inmanta logging level
    :param cli: True if the logs will be outputted to the CLI.
    :return: python log level
    """
    # maximum of 4 v's
    if inmanta_log_level.isdigit() and int(inmanta_log_level) > 4:
        inmanta_log_level = "4"
    # The minimal log level on the CLI is always WARNING
    if cli and (inmanta_log_level == "ERROR" or (inmanta_log_level.isdigit() and int(inmanta_log_level) < 1)):
        inmanta_log_level = "WARNING"
    # Converts the Inmanta log level to the Python log level
    python_log_level = log_levels[inmanta_log_level]
    return python_log_level


class LoggingConfigSource(abc.ABC):
    """
    A class that indicates where the logging configuration comes from.
    """

    @abc.abstractmethod
    def read_logging_config(self, context: Mapping[str, str]) -> dict[str, object]:
        """
        Read the logging config from this LoggingConfigSource and return it in dictionary form.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def is_template(self) -> bool:
        """
        Return True iff this LoggingConfigSource needs to be considered as a template.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def source(self) -> str:
        """
        Return a string representation that indicates the source of the config. This is used in error reporting.
        """
        raise NotImplementedError()

    def render_logging_config_template(self, template: str, context: Mapping[str, str]) -> str:
        """
        This method fills in the template variables present in the given logging configuration template.

        :param template: The logging configuration template
        :param context: The context variables that should be used to populate the template
        """
        try:
            return template.format(**context)
        except KeyError as e:
            all_keys = ", ".join(context.keys())
            # Not very good exception
            raise Exception(
                f"The logging configuration template from {self.source()} refers to context variable {str(e)}, "
                f"but this variable is not available. The context is limited to {all_keys}"
            )


class LoggingConfigFromFile(LoggingConfigSource):
    """
    A LoggingConfig present in a file.
    """

    def __init__(self, file_name: str) -> None:
        self.file_name = os.path.abspath(file_name)

    def read_logging_config(self, context: Mapping[str, str]) -> dict[str, object]:
        try:
            with open(self.file_name, "r") as fh:
                logging_config_as_str = fh.read()
        except FileNotFoundError:
            raise Exception(f"Logging config file {self.file_name} doesn't exist.")
        except Exception:
            raise Exception(f"Failed to read logging config file from {self.file_name}.")

        if self.is_template():
            logging_config_as_str = self.render_logging_config_template(template=logging_config_as_str, context=context)

        try:
            return yaml.safe_load(logging_config_as_str)
        except Exception:
            raise Exception(f"Failed to parse logging config file from {self.file_name} as yaml.")

    def is_template(self) -> bool:
        return self.file_name.endswith(".tmpl")

    def source(self) -> str:
        return f"file {self.file_name}"


class LoggingConfigFromEnvVar(LoggingConfigSource):
    """
    A logging config present in an environment variable (not a reference to a file).
    """

    def __init__(self, env_var_name: str) -> None:
        self.env_var_name = env_var_name

    def read_logging_config(self, context: Mapping[str, str]) -> dict[str, object]:
        logging_config_as_str = os.getenv(self.env_var_name, None)
        if logging_config_as_str is None:
            raise Exception(f"Environment variable {self.env_var_name} not found.")

        if self.is_template():
            logging_config_as_str = self.render_logging_config_template(template=logging_config_as_str, context=context)

        try:
            return yaml.safe_load(logging_config_as_str)
        except Exception:
            raise Exception(f"Failed to parse logging config from environment variable {self.env_var_name} as yaml.")

    def is_template(self) -> bool:
        return self.env_var_name.endswith("_TMPL")

    def source(self) -> str:
        return f"environment variable {self.env_var_name}"


@stable_api
class InmantaLoggerConfig:
    """
    This class is the entry-point for configuring the Python logging framework.

    Usage:
    To use this class, you first need to call the `get_instance`. This method takes a `stream` argument
    that specifies where the log messages should be sent to. If no `stream` is provided,
    the log messages will be sent to standard output.

    You can then call the `apply_options` method to configure the logging options.

    The setup is not done in one step as we want logs for the cmd_parser, which will provide the options needed to configure
    the 'final' logger with `apply_options`.
    """

    _instance: Optional["InmantaLoggerConfig"] = None

    def __init__(self, stream: TextIO = sys.stdout, no_install: bool = False) -> None:
        """
        Set up the logging handler for Inmanta

        :param stream: The TextIO stream where the logs will be sent to.
        :param no_install: True iff we don't want to install (apply) the logging config to the Python logging framework.
                           This parameter is set to True by the print-default-logging-config tool.
        """
        self.no_install = no_install

        log_config: FullLoggingConfig = LoggingConfigBuilder().get_bootstrap_logging_config(stream)
        self._stream = stream
        self._handlers: Sequence[logging.Handler] = self._apply_logging_config(log_config)

        self._loaded_config: FullLoggingConfig | None = None
        if logfire_enabled:
            logging.root.addHandler(LogfireLoggingHandler())

        # cache for original startup config
        self._options_applied: Options | None = None
        self._component: str | None = None
        self._context: Mapping[str, str] | None = None

        self.logging_config_source: LoggingConfigSource | None = None

    @classmethod
    def get_current_instance(cls) -> "InmantaLoggerConfig":
        """
        Obtain the InmantaLoggerConfig singleton. This method assumes that an InmantaLoggerConfig was already initialized
        using the `get_instance()` method.
        """
        if not cls._instance:
            raise Exception("InmantaLoggerConfig was not yet initialized. Call get_instance() first.")
        return cls._instance

    @classmethod
    @stable_api
    def get_instance(cls, stream: TextIO = sys.stdout) -> "InmantaLoggerConfig":
        """
        This method should be used to obtain an instance of this class, because this class is a singleton.

        :param stream: The stream to send log messages to. Default is standard output (sys.stdout)
        """
        if cls._instance:
            if not cls._instance._handlers:
                raise Exception("No handlers found.")
            if not isinstance(cls._instance._handlers[0], logging.StreamHandler):
                raise Exception("Instance already exists with a different handler")
            elif isinstance(cls._instance._handlers[0], logging.StreamHandler) and cls._instance._handlers[0].stream != stream:
                raise Exception("Instance already exists with a different stream")
        else:
            cls._instance = cls(stream)
        return cls._instance

    @classmethod
    @stable_api
    def clean_instance(cls, root_handlers_to_remove: Optional[Sequence[logging.Handler]] = None) -> None:
        """
        This method should be used to clean up an instance of this class.

        By default, this method removes and closes all root handlers from the logging framework. If the
        root_handlers_to_remove argument is not None, only the provided root handlers will be removed and closed.
        """
        logging.shutdown()
        cls._instance = None

    def _get_content_env_var_for_component(self, component_name: str | None) -> str | None:
        """
        Returns the name of the environment variable that contains the content of the logging configuration
        for the given component. This can be either the regular (`*_CONTENT`) or the template-based (`*_TMPL`)
        environment variable. None is returned if none of the environment variables was populated.
        """
        option = component_log_configs[component_name] if component_name is not None else logging_config
        base_env_var_name = option.get_environment_variable()
        content_env_var_names = [f"{base_env_var_name}_CONTENT", f"{base_env_var_name}_TMPL"]
        env_vars_set_by_user = set(content_env_var_names) & os.environ.keys()
        if not env_vars_set_by_user:
            return None
        if len(env_vars_set_by_user) > 1:
            LOGGER.warning(
                "Environment variables %s and %s are set simultaneously. Using %s.",
                content_env_var_names[0],
                content_env_var_names[1],
                content_env_var_names[0],
            )
            return content_env_var_names[0]
        return env_vars_set_by_user.pop()

    def _get_logging_config_source_for_component(self, component_name: str | None) -> LoggingConfigSource | None:
        # Check if one of the environment variables, that contain the content of the logging config, are set.
        content_env_var: str | None = self._get_content_env_var_for_component(component_name=component_name)
        if content_env_var is not None:
            return LoggingConfigFromEnvVar(env_var_name=content_env_var)
        # Check if the configuration option for the given component is set that references a logging config file.
        config_option: Option[str | None] = (
            component_log_configs[component_name] if component_name is not None else logging_config
        )
        # The logging config of a component defaults to the general config.logging_config config option.
        # As such, we ignore the default here when a component_name is provided.
        file_name: str | None = config_option.get()
        if file_name is not None:
            return LoggingConfigFromFile(file_name=file_name)
        # No logging configuration was found for the given component.
        return None

    def _get_logging_config_source(self, options: Options, component: str | None = None) -> LoggingConfigSource:
        """
        This method returns the source of the logging config that should be loaded according to the precedence rules
        for the logging configuration. The following precedence rules are taken into account (lower number higher precedence):

          1. The --logging-config CLI option.
          2. The INMANTA_LOGGING_<COMPONENT>_CONTENT and INMANTA_LOGGING_<COMPONENT>_TMPL environment variables.
          3. The INMANTA_LOGGING_<COMPONENT> environment variable.
          4. The component specific log config option.
          5. The INMANTA_CONFIG_LOGGING_CONFIG_CONTENT and INMANTA_CONFIG_LOGGING_CONFIG_TMPL environment variables.
          6. The INMANTA_CONFIG_LOGGING_CONFIG environment variable.
          7. The config.logging_config option in the config files.

          If the user didn't specify any log config, using any of the above-mentioned methods, a NoLoggingConfigFound
          exception is raised.

        :param options: The CLI options.
        :param component: The name of the component being executed.
        """
        # Check --logging-config CLI option.
        if options.logging_config:
            return LoggingConfigFromFile(file_name=options.logging_config)
        # Check component-specific logging config options.
        if component is not None:
            source = self._get_logging_config_source_for_component(component_name=component)
            if source is not None:
                return source
        # Check component-independent logging config options.
        source = self._get_logging_config_source_for_component(component_name=None)
        if source is not None:
            return source
        raise NoLoggingConfigFound()

    @stable_api
    def apply_options(self, options: Options, component: str | None = None, context: Mapping[str, str] | None = None) -> None:
        """
        Apply the logging options to the current handler. A handler should have been created before

        :param options: The Option object coming from the command line. This function uses the following
            attributes: log_file, log_file_level, verbose, timed
        :param component: The component to configure (e.g. server, scheduler, compiler).
        Used to select which config file option to use (logging.component)
        :param context: context variables to use if the config file is a template
        """

        def user_defined_options() -> str:
            """
            Returns a string with the options (excluding "--verbose" and "--logging-config") that were set by the user
            """
            args_to_cli = {
                "log_file": "--log-file",
                "log_file_level": "--log-file-level",
            }
            flags_to_cli = {
                "timed": "--timed-logs",
                "keep_logger_names": "--keep-logger-names",
            }
            ignored_options_list = []
            for key, value in {**args_to_cli, **flags_to_cli}.items():
                if key not in options:
                    continue
                ignored_options_list.append(f"{value} {getattr(options, key)}" if key in args_to_cli else value)

            return ", ".join(ignored_options_list)

        if self._options_applied:
            raise Exception("Options can only be applied once to a handler.")
        if context is None:
            context = {}

        try:
            self.logging_config_source = self._get_logging_config_source(options, component)
        except NoLoggingConfigFound:
            # No logging config was defined by the user. Compose the logging config from the old CLI options.
            self._apply_logging_config_from_options(options, component, context)
        else:
            # A logging config was defined by the user, apply it to the logging framework.
            logging_config_as_dct = self.logging_config_source.read_logging_config(context)
            self._apply_logging_config_from_dict(logging_config_as_dct)
            # Take into account the verbosity flag on the CLI.
            if options.verbose != 0:
                self.force_cli(convert_inmanta_log_level(str(options.verbose), cli=True))
            ignored_options = user_defined_options()
            if ignored_options:
                LOGGER.warning(
                    "Ignoring the following options: %s. Using logging config from %s",
                    ignored_options,
                    self.logging_config_source.source(),
                )

        self._options_applied = options
        self._component = component
        self._context = context

    def extend_config(self, extenders: "list[LoggingConfigBuilderExtension]") -> FullLoggingConfig:
        """
        Second stage loading: add config extensions
        """
        if not self._options_applied:
            raise Exception("Extenders can only be added after loading the initial config")
        assert self._context is not None  # make mypy happy
        assert self._loaded_config is not None  # make mypy happy

        if not extenders:
            # No extensions, easy
            return self._loaded_config

        if self.logging_config_source:
            # A logging config was defined, no extenders needed
            return self._loaded_config

        config = self._loaded_config
        assert config is not None  # make mypy happy
        for extender in extenders:
            config = extender.get_logging_config_from_options(
                self._stream, self._options_applied, self._component, self._context, config
            )
        self._handlers = self._apply_logging_config(config)
        return config

    def force_cli(self, python_log_level: int) -> None:
        """Ensure a cli logger is attached at the given level"""
        config_builder = LoggingConfigBuilder()
        logger_config = config_builder.get_bootstrap_logging_config(python_log_level=python_log_level)
        self.add_extension_config(logger_config)

    def _apply_logging_config_from_dict(self, dict_config: dict[str, object]) -> None:
        """
        Apply the given logging config dictionary.
        """
        handlers_before = list(logging.root.handlers)
        try:
            logging.config.dictConfig(dict_config)
        except Exception:
            raise Exception(f"Failed to apply the logging config defined in {dict_config}.")
        self._handlers = [handler for handler in logging.root.handlers if handler not in handlers_before]

        def as_dict(inp: dict[str, object], key: str) -> dict[str, object]:
            root = inp.get(key, {})
            if not isinstance(root, dict):
                raise Exception(f"{key} entry should be a dict, got {root}")
            return root

        root = as_dict(dict_config, "root")
        # Build config for later merges
        self._loaded_config = FullLoggingConfig(
            formatters=as_dict(dict_config, "formatters"),
            handlers=as_dict(dict_config, "handlers"),
            loggers=as_dict(dict_config, "loggers"),
            root_handlers=root.get("handlers", []),
            root_log_level=root.get("level", None),
        )

    def _apply_logging_config_from_options(self, options: Options, component: str | None, context: Mapping[str, str]) -> None:
        """
        Apply the logging configuration as defined by the CLI options when the --logging-config option is not set.
        """
        config_builder = LoggingConfigBuilder()
        logging_config: FullLoggingConfig = config_builder.get_logging_config_from_options(
            self._stream, options, component, context
        )
        self._handlers = self._apply_logging_config(logging_config)

    def _apply_logging_config(self, logging_config: FullLoggingConfig) -> Sequence[logging.Handler]:
        """
        Apply the given logging_config as the current configuration of the logging system.

        This method assume that the given config defines a single root handler.
        """
        handlers_before = list(logging.root.handlers)

        if not self.no_install:
            logging_config.apply_config()
        self._loaded_config = logging_config
        return [handler for handler in logging.root.handlers if handler not in handlers_before]

    @stable_api
    def add_extension_config(self, logging_config: LoggingConfigExtension) -> None:
        """
        Register the default logging config for a certain extension.
        """
        assert self._loaded_config is not None
        complete_config = self._loaded_config.join(logging_config, allow_overwrite=True)
        self._apply_logging_config(complete_config)


@stable_api
class MultiLineFormatter(colorlog.ColoredFormatter):
    """
    Formatter for multi-line log records.

    This class extends the `colorlog.ColoredFormatter` class to provide a custom formatting method for log records that
    span multiple lines.
    """

    inmanta_plugin_pkg_regex = re.compile(r"^inmanta_plugins\.(?P<module_name>[^.]+)")
    # Regex that extracts the name of the module from a fully qualified import of a Python
    # module inside an Inmanta module.

    def __init__(
        self,
        fmt: Optional[str] = None,
        *,
        # keep interface minimal: only include fields we actually use
        log_colors: Optional[LogColors] = None,
        reset: bool = True,
        no_color: bool = False,
        keep_logger_names: bool = True,
    ):
        """
        Initialize a new `MultiLineFormatter` instance.

        :param fmt: Optional string specifying the log record format.
        :param log_colors: Optional `LogColors` object mapping log level names to color codes.
        :param reset: Boolean indicating whether to reset terminal colors at the end of each log record.
        :param no_color: Boolean indicating whether to disable colors in the output.
        :param keep_logger_names: Display the log messages using the name of the logger that created the log message,
                                  instead of the component of the compiler that was executing while the log record was created
                                  or the name of the module that created the log message.
        """
        super().__init__(fmt, log_colors=log_colors, reset=reset, no_color=no_color)
        self.fmt = fmt
        self._keep_logger_names = keep_logger_names

    def get_header_length(self, record: logging.LogRecord) -> int:
        """
        Get the header length of a given log record.

        :param record: The `logging.LogRecord` object for which to calculate the header length.
        :return: The length of the header in the log record, without color codes.
        """
        # to get the length of the header we want to get the header without the color codes
        formatter = colorlog.ColoredFormatter(
            fmt=self.fmt,
            log_colors=self.log_colors,
            reset=False,
            no_color=True,
        )
        header = formatter.format(
            logging.LogRecord(
                record.name,
                record.levelno,
                record.pathname,
                record.lineno,
                "",
                (),
                None,
            )
        )
        return len(header)

    def format(self, record: logging.LogRecord) -> str:
        """
        Format a log record with added indentation.

        :param record: The `logging.LogRecord` object to format.
        :return: The formatted log record as a string.
        """
        if not self._keep_logger_names:
            record = self._wrap_record(record)
        indent: str = " " * self.get_header_length(record)
        head, *tail = super().format(record).splitlines(True)
        return head + "".join(indent + line for line in tail)

    def _wrap_record(self, record: logging.LogRecord) -> logging.LogRecord:
        """
        Rename the logger for the given log record. This only affects logs generated by
        the inmanta app. This is both used to display shorter names and to group records generated
        by different modules under a common name.
        """
        old_name = record.name
        new_name = self._get_logger_name_for(old_name)
        if old_name == new_name:
            return record
        attributes = dict(record.__dict__)
        attributes["name"] = new_name
        return logging.makeLogRecord(attributes)

    def _get_logger_name_for(self, logger_name: str) -> str:
        """
        Returns the logger name that should be used in the log record.

        :attr logger_name: The name of the logger that was used to create the log record.
        """
        if not logger_name.startswith("inmanta"):
            # This is a log record from a third-party library. Don't adjust the logger name.
            return logger_name
        if logger_name == "inmanta.pip":
            # Log record created by a pip subprocess started by the inmanta.
            return "pip"
        match: Optional[re.Match[str]] = self.inmanta_plugin_pkg_regex.match(logger_name)
        if match:
            # Log record created by an Inmanta module.
            return match.groupdict()["module_name"]
        elif "export" in logger_name:
            return "exporter"
        elif "protocol" in logger_name:
            return "exporter"
        else:
            # Log record created by Inmanta code.
            return "compiler"


@stable_api
class ParametrizedFileHandler(logging.Handler):
    """
    A file handler that writes to different files based on the last part of the logger name.
    """

    def __init__(self, name_parent_logger: str, log_file_template: str) -> None:
        """
        :param name_parent_logger: The log records created by children of this logger will be written to file by this handler.
                                   This handler will ignore log records created by any other logger.
        :param log_file_template: A template for the path to the log file. This is an f-string that holds the parameter
                                  `child_logger_name`. This part will be replaced by the name of the direct child of
                                  self.name_parent_logger that belongs to the logger that created the record.
        """
        super().__init__()
        self.name_parent_logger = name_parent_logger
        self.log_file_template: str = log_file_template
        self.child_handlers: dict[str, handlers.WatchedFileHandler] = {}

    def flush(self) -> None:
        for child in self.child_handlers.values():
            child.flush()

    def setFormatter(self, fmt: Optional[logging.Formatter]) -> None:
        super().setFormatter(fmt)
        for child in self.child_handlers.values():
            child.setFormatter(fmt)

    def setLevel(self, level: str | int) -> None:
        super().setLevel(level)
        for child in self.child_handlers.values():
            child.setLevel(level)

    def close(self) -> None:
        for child in self.child_handlers.values():
            child.close()
        self.child_handlers = {}
        super().close()

    def emit(self, record: logging.LogRecord) -> None:
        if not record.name.startswith(f"{self.name_parent_logger}."):
            return
        logger_name_without_parent_prefix = record.name.removeprefix(f"{self.name_parent_logger}.")
        logger_name_direct_child_of_parent_logger = logger_name_without_parent_prefix.split(".")[0]
        path_logfile = self.log_file_template.format(child_logger_name=logger_name_direct_child_of_parent_logger)
        if path_logfile not in self.child_handlers:
            try:
                handler = logging.handlers.WatchedFileHandler(filename=path_logfile, mode="a+")
            except FileNotFoundError:
                log_dir = os.path.dirname(self.log_file_template)
                if not os.path.exists(log_dir):
                    LOGGER.warning(
                        "Cannot write to resource action log, because directory %s doesn't exist.",
                        path_logfile,
                    )
                else:
                    LOGGER.exception("Failed to write to resource action log file.")
                return
            handler.setFormatter(self.formatter)
            handler.setLevel(self.level)
            self.child_handlers[path_logfile] = handler
        self.child_handlers[path_logfile].emit(record)


class TornadoDebugLogHandler(logging.Handler):
    """
    A custom log handler for Tornados 'max_clients limit reached' debug logs.
    """

    def __init__(self, level: int = logging.NOTSET) -> None:
        super().__init__(level)
        self.logger = logging.getLogger("inmanta.protocol.endpoints")

    def emit(self, record: logging.LogRecord) -> None:
        if (
            record.levelno == logging.DEBUG
            and record.name.startswith("tornado.general")
            and record.msg.startswith("max_clients limit reached")
        ):
            self.logger.warning(record.msg)  # Log Tornado log as inmanta warnings



"""
Copyright 2020 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import inspect
import itertools
import json
import logging
import re
from collections.abc import Sequence
from typing import Callable, Optional, Union

from pydantic import ConfigDict
from typing_inspect import get_args, get_origin, is_generic_type

from inmanta import util
from inmanta.const import INMANTA_MT_HEADER
from inmanta.data.model import BaseModel
from inmanta.protocol.common import ArgOption, MethodProperties, ReturnValue, UrlMethod
from inmanta.protocol.openapi.model import (
    CodeSample,
    Components,
    Header,
    Info,
    MediaType,
    OpenAPI,
    OpenApiDataTypes,
    Operation,
    Parameter,
    ParameterType,
    PathItem,
    Reference,
    RequestBody,
    Response,
    Schema,
    Scope,
    SecurityScheme,
    SecuritySchemeName,
    Server,
)
from inmanta.server import config
from inmanta.server.extensions import FeatureManager
from inmanta.types import ReturnTypes

LOGGER = logging.getLogger(__name__)


def openapi_json_encoder(o: object) -> Union[ReturnTypes, util.JSONSerializable]:
    if isinstance(o, BaseModel):
        return o.dict(by_alias=True, exclude_none=True)
    return util.api_boundary_json_encoder(o)


class OpenApiConverter:
    """
    Extracts API information for the OpenAPI definition from the server
    """

    def __init__(self, global_url_map: dict[str, dict[str, UrlMethod]], feature_manager: FeatureManager):
        self.global_url_map = global_url_map
        self.feature_manager = feature_manager
        self.type_converter = OpenApiTypeConverter()
        self.arg_option_handler = ArgOptionHandler(self.type_converter)

    def _collect_server_information(self) -> list[Server]:
        bind_port = config.server_bind_port.get()
        server_address = config.server_address.get()
        protocol = "https" if config.ssl_enabled() else "http"
        return [
            Server(
                url=f"{protocol}://{server_address}:{bind_port}/",
            ),
            Server(url="/"),  # Generic server, working even if the swagger is behind a reverse proxy
        ]

    def _get_inmanta_version(self) -> Optional[str]:
        metadata = self.feature_manager.get_product_metadata()
        return metadata.version

    def generate_openapi_definition(self) -> OpenAPI:
        version = self._get_inmanta_version()
        info = Info(title="Inmanta Service Orchestrator", version=version if version else "")
        servers = self._collect_server_information()
        paths = {}
        for path, methods in self.global_url_map.items():
            api_methods = self._filter_api_methods(methods)
            if len(api_methods) > 0:
                url_method: UrlMethod = next(iter(methods.values()))
                parameterized_path: str = url_method.get_path()
                path_item = self._extract_operations_from_methods(api_methods, parameterized_path)
                paths[parameterized_path] = path_item
        security: list[dict[SecuritySchemeName, list[Scope]]] | None = (
            # Set the security scheme globally for all endpoints.
            [{schema_name: []} for schema_name in self.type_converter.components.securitySchemes]
            if self.type_converter.components.securitySchemes
            else None
        )
        return OpenAPI(
            openapi="3.0.2",
            info=info,
            paths=paths,
            servers=servers,
            components=self.type_converter.components,
            security=security,
        )

    def _filter_api_methods(self, methods: dict[str, UrlMethod]) -> dict[str, UrlMethod]:
        return {
            method_name: url_method
            for method_name, url_method in methods.items()
            if "api" in url_method.properties.client_types
        }

    def _extract_operations_from_methods(self, api_methods: dict[str, UrlMethod], path: str) -> PathItem:
        path_item = PathItem()
        for http_method_name, url_method in api_methods.items():
            operation_handler = OperationHandler(self.type_converter, self.arg_option_handler)
            operation = operation_handler.handle_method(url_method, path)
            path_item.__setattr__(http_method_name.lower(), operation)
        return path_item

    def generate_openapi_json(self) -> str:
        openapi = self.generate_openapi_definition()

        return json.dumps(openapi, default=openapi_json_encoder)

    def generate_swagger_html(self) -> str:
        return self.get_swagger_html(self.generate_openapi_json())

    def get_swagger_html(self, openapi_spec: str) -> str:
        template = f"""
        <!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <script src="//unpkg.com/swagger-ui-dist@^4.11.1/swagger-ui-bundle.js"></script>
        <link rel="stylesheet" href="//unpkg.com/swagger-ui-dist@^4.11.1/swagger-ui.css" />
        <title>Inmanta Service Orchestrator API</title>
    </head>
    <body>
        <div id="swagger-ui"></div>
        <script>
            window.onload = function() {{
              SwaggerUIBundle({{
                spec: {openapi_spec},
                dom_id: '#swagger-ui',
                presets: [
                  SwaggerUIBundle.presets.apis
                ],
              }})
            }}
        </script>
    </body>
    </html>
        """

        return template


class OpenApiTypeConverter:
    """
    Lookup for OpenAPI types corresponding to python types
    """

    def __init__(self) -> None:
        self.components = Components(schemas={}, securitySchemes=self._get_security_schemes())
        self._pydantic_ref_key = "$defs"
        self.openapi_ref_prefix = "#/components/schemas/"
        self._ref_regex = re.compile(re.escape(self.openapi_ref_prefix) + r"(.*)")

    def _get_security_schemes(self) -> dict[SecuritySchemeName, SecurityScheme] | None:
        if config.server_enable_auth.get():
            return {
                "bearerAuth": SecurityScheme(
                    type="http",
                    scheme="bearer",
                    bearerFormat="JWT",
                )
            }
        else:
            return None

    def get_openapi_type_of_parameter(self, parameter_type: inspect.Parameter) -> Schema:
        schema = self.get_openapi_type(parameter_type.annotation)
        if parameter_type.default is not inspect.Parameter.empty:
            schema.default = parameter_type.default
        return schema

    def _handle_pydantic_model(self, type_annotation: BaseModel, by_alias: bool = True) -> Schema:
        schema = type_annotation.model_json_schema(by_alias=by_alias, ref_template=f"{self.openapi_ref_prefix}{{model}}")
        # pydantic.BaseModel.model_json_schema() stores the model (and sub-model) definitions at #/$defs/{model}
        # (pydantic.main.DEFAULT_REF_TEMPLATE) but OpenAPI requires them to be placed at "#/components/schemas/"
        # The ref_template tells pydantic to update the references (the referring side) but not the schema location (the
        # referred side). So we need to perform that translation manually.
        if self._pydantic_ref_key in schema.keys():
            definitions: dict[str, dict[str, object]] = schema.pop(self._pydantic_ref_key)
            if self.components.schemas is not None:
                for key, definition in definitions.items():
                    patched_definition = self._handle_pydantic_model_recursively(definition)
                    self.components.schemas[key] = Schema(**patched_definition)

        schema = self._handle_pydantic_model_recursively(schema)
        return Schema(**schema)

    def _handle_pydantic_model_recursively(self, definition: dict[str, object]) -> dict[str, object]:
        """
        The json-schema is not 1-to-1 compatible with the OpenAPI schema. This method patches the
        given json-schema as such that it becomes compatible with the OpenAPI schema.
        """
        # Recursive calls
        if definition.get("properties") is not None:
            properties = definition["properties"]
            for property_name in list(properties.keys()):
                properties[property_name] = self._handle_pydantic_model_recursively(properties[property_name])
        if definition.get("additionalProperties") is not None and isinstance(definition["additionalProperties"], dict):
            definition["additionalProperties"] = self._handle_pydantic_model_recursively(definition["additionalProperties"])
        if definition.get("items") is not None:
            definition["items"] = self._handle_pydantic_model_recursively(definition["items"])
        for attr_name in ["anyOf", "allOf", "oneOf"]:
            if definition.get(attr_name) is not None:
                for i in range(len(definition[attr_name])):
                    definition[attr_name][i] = self._handle_pydantic_model_recursively(definition[attr_name][i])

        definition = self._convert_const_to_enum(definition)
        definition = self._add_type_field_to_enum_value(definition)
        return definition

    def _convert_const_to_enum(self, definition: dict[str, object]) -> dict[str, object]:
        """
        The 'const' field, which is part of the JSON schema, is not part of the OpenAPI schema.
        We convert the 'const' field into an 'enum' field here to make sure the constant value
        of the parameter is available in the OpenAPI schema.
        """
        if definition.get("enum") is not None:
            # The enum field is already populated
            return definition
        if definition.get("const") is not None:
            definition["enum"] = [str(definition["const"])]
            del definition["const"]
        return definition

    def _add_type_field_to_enum_value(self, definition: dict[str, object]) -> dict[str, object]:
        """
        When pydantic converts a Python Enum type to its corresponding json schema, it doesn't
        populate the type field. This way the rendered API documentation doesn't include all possible
        enum values. This method makes sure that the type attribute of an enum is always populated.
        """
        if definition.get("enum") is not None:
            if not definition.get("type"):
                # Convert Python type to a type known by OpenAPI
                if all(isinstance(e, bool) for e in definition["enum"]):
                    definition["type"] = OpenApiDataTypes.BOOLEAN.value
                elif all(isinstance(e, int) for e in definition["enum"]):
                    definition["type"] = OpenApiDataTypes.INTEGER.value
                elif all(isinstance(e, float) or isinstance(e, int) for e in definition["enum"]):
                    definition["type"] = OpenApiDataTypes.NUMBER.value
                else:
                    definition["type"] = OpenApiDataTypes.STRING.value
            # OpenAPI expects enum values to be strings
            definition["enum"] = [str(e) for e in definition["enum"]]
        return definition

    def get_openapi_type(self, type_annotation: type) -> Schema:
        class Sub(BaseModel):
            the_field: type_annotation
            model_config = ConfigDict(arbitrary_types_allowed=True)

        pydantic_result = self._handle_pydantic_model(Sub).properties["the_field"]
        pydantic_result.title = None
        return pydantic_result

    def resolve_reference(self, reference: str) -> Optional[Schema]:
        """
        Get a schema from its reference, if this schema exists.  If it doesn't exist, None is
        returned instead.

        :param reference: The reference to the schema
        """
        ref_match = self._ref_regex.match(reference)
        if not ref_match:
            return None

        ref_key = ref_match.group(1)
        return self.components.schemas[ref_key]


class ArgOptionHandler:
    """
    Extracts header, response header and path parameter information from ArgOptions
    """

    def __init__(self, type_converter: OpenApiTypeConverter) -> None:
        self.type_converter = type_converter

    def extract_parameters_from_arg_options(
        self, method_properties: MethodProperties, function_parameters: dict[str, inspect.Parameter]
    ) -> list[Parameter]:
        result: list[Parameter] = []
        for option_name, option in method_properties.arg_options.items():
            param = function_parameters[option_name]
            param_schema = self.type_converter.get_openapi_type_of_parameter(param)
            param_description = method_properties.get_description_for_param(option_name)
            if option.header:
                result.append(
                    Parameter(in_=ParameterType.header, name=option.header, schema_=param_schema, description=param_description)
                )
        return result

    def extract_response_headers_from_arg_options(
        self, arg_options: dict[str, ArgOption]
    ) -> Optional[dict[str, Union[Header, Reference]]]:
        headers: dict[str, Union[Header, Reference]] = {}
        for option_name, option in arg_options.items():
            if option.header and option.reply_header:
                headers[option.header] = Header(
                    description=f"The value of the request header {option.header}", schema=Schema(type="string")
                )
        return headers if headers else None


class FunctionParameterHandler:
    """
    Creates OpenAPI Parameters and RequestBody items based on the handler function parameters
    """

    def __init__(
        self,
        type_converter: OpenApiTypeConverter,
        arg_option_handler: ArgOptionHandler,
        path: str,
        method_properties: MethodProperties,
    ):
        self.type_converter = type_converter
        self.arg_option_handler = arg_option_handler
        self.path = path
        self.method_properties = method_properties

        # Get the parameters of the handler function
        self.all_params_dct: dict[str, inspect.Parameter] = self._extract_function_parameters(method_properties.function)
        self.path_params: dict[str, inspect.Parameter] = {}
        self.header_params: dict[str, inspect.Parameter] = {}
        self.non_path_and_non_header_params: dict[str, inspect.Parameter] = {}

        for param_name, param in self.all_params_dct.items():
            if f"<{param_name}>" in self.path:
                self.path_params[param_name] = param
            elif (
                param_name in self.method_properties.arg_options.keys()
                and self.method_properties.arg_options[param_name].header is not None
            ):
                self.header_params[param_name] = param
            else:
                self.non_path_and_non_header_params[param_name] = param

    # Parameters

    def _extract_function_parameters(self, url_method_function: Callable) -> dict[str, inspect.Parameter]:
        function_parameters = {
            parameter_name: parameter_type
            for parameter_name, parameter_type in inspect.signature(url_method_function).parameters.items()
        }
        return function_parameters

    def get_parameters(self) -> list[Parameter]:
        result = self._convert_header_and_path_params()
        if self.method_properties.operation not in ["POST", "PUT", "PATCH"]:
            result.extend(self._convert_function_params_to_query_params())
        return result

    def _convert_header_and_path_params(self) -> list[Parameter]:
        arg_options_params: list[Parameter] = self.arg_option_handler.extract_parameters_from_arg_options(
            self.method_properties, self.all_params_dct
        )
        path_params: list[Parameter] = self._convert_path_params_to_openapi()
        return arg_options_params + path_params

    def _convert_path_params_to_openapi(self) -> list[Parameter]:
        parameters: list[Union[Parameter, Reference]] = []
        for parameter_name, parameter_type in self.path_params.items():
            type_description = self.type_converter.get_openapi_type_of_parameter(parameter_type)
            param_description = self.method_properties.get_description_for_param(parameter_name)
            parameters.append(
                Parameter(
                    name=parameter_name,
                    in_=ParameterType.path,
                    required=True,
                    schema_=type_description,
                    description=param_description,
                )
            )
        return parameters

    def _convert_function_params_to_query_params(self) -> list[Parameter]:
        parameters: list[Parameter] = []
        for parameter_name, parameter_type in self.non_path_and_non_header_params.items():
            type_description = self.type_converter.get_openapi_type_of_parameter(parameter_type)
            param_description = self.method_properties.get_description_for_param(parameter_name)
            parameters.append(
                Parameter(name=parameter_name, in_=ParameterType.query, schema_=type_description, description=param_description)
            )
        return parameters

    # Request body

    def convert_request_body(self) -> RequestBody:
        properties = self._convert_function_params_to_openapi_request_body_properties()
        return self._build_json_request_body(properties)

    def _convert_function_params_to_openapi_request_body_properties(self) -> dict[str, Schema]:
        properties = {}
        for parameter_name, parameter_type in self.non_path_and_non_header_params.items():
            type_description = self.type_converter.get_openapi_type_of_parameter(parameter_type)
            properties[parameter_name] = type_description
        return properties

    def _get_request_body_description(self) -> str:
        """
        OpenAPI supports only a single description field for the full request body. As such,
        this method return the description of the request body in CommonMark syntax to create
        an itemization of the different parameters in the request body.
        """
        result = ""
        for param_name in sorted(self.non_path_and_non_header_params.keys()):
            description = self.method_properties.get_description_for_param(param_name)
            if description is not None:
                result += f"* **{param_name}:** {description}\n"
            else:
                result += f"* **{param_name}:**\n"
        return result

    def _build_json_request_body(self, properties: dict) -> RequestBody:
        request_body = RequestBody(
            required=True,
            content={"application/json": MediaType(schema=Schema(type="object", properties=properties))},
            description=self._get_request_body_description(),
        )
        return request_body


class OperationHandler:
    """
    Builds an OpenAPI Operation object from an Inmanta UrlMethod
    """

    def __init__(self, type_converter: OpenApiTypeConverter, arg_option_handler: ArgOptionHandler):
        self.type_converter = type_converter
        self.arg_option_handler = arg_option_handler

    def handle_method(self, url_method: UrlMethod, path: str) -> Operation:
        function_parameter_handler = FunctionParameterHandler(
            self.type_converter, self.arg_option_handler, path, url_method.properties
        )
        parameters = function_parameter_handler.get_parameters()
        responses = self._build_responses(url_method.properties)

        request_body_parameters: list[str] = []

        if url_method.get_operation() in ["POST", "PUT", "PATCH"]:
            extra_params = {"requestBody": function_parameter_handler.convert_request_body()}
            try:
                request_body_parameters = list(
                    extra_params["requestBody"].content["application/json"].schema_.properties.keys()
                )
            except (KeyError, AttributeError):
                pass
        else:
            extra_params = {}

        tags = self._get_tags_of_operation(url_method)
        code_samples = self._generate_code_sample(
            method_name=url_method.method_name, parameters=parameters, request_body_parameters=request_body_parameters
        )
        return Operation(
            responses=responses,
            operationId=url_method.method_name,
            parameters=(parameters if len(parameters) else None),
            code_samples=code_samples,
            summary=url_method.short_method_description,
            description=url_method.long_method_description,
            tags=tags,
            **extra_params,
        )

    def _generate_code_sample(
        self, method_name: str, parameters: Sequence[Parameter], request_body_parameters: Sequence[str]
    ) -> list[CodeSample]:
        """
        Generate a python code sample for the given method and parameters. The returned sample is wrapped in a list to match
        the type expected by redoc
        (see https://redocly.com/docs-legacy/api-reference-docs/specification-extensions/x-code-samples#usage)
        """
        tab_padding = " " * 4
        all_parameters = list(
            itertools.chain(
                [param.name if param.name != INMANTA_MT_HEADER else "tid" for param in parameters], request_body_parameters
            )
        )
        arguments = f"\n{',\n'.join(f"{tab_padding}{param}=..." for param in all_parameters)}\n" if all_parameters else ""
        source = (
            "client = inmanta.protocol.endpoints.Client(name='api', timeout=120)\n"
            f"result = await client.{method_name}({arguments})"
        )
        code_samples = CodeSample(lang="Python", source=source)
        return [code_samples]

    def _get_tags_of_operation(self, url_method: UrlMethod) -> Optional[list[str]]:
        if url_method.endpoint is not None:
            if hasattr(url_method.endpoint, "_name"):
                return [url_method.endpoint._name]
            else:
                return [url_method.endpoint.__class__.__name__]
        else:
            return None

    def _build_responses(self, url_method_properties: MethodProperties) -> dict[str, Response]:
        result: dict[str, Response] = {}
        status_code_to_description_map = url_method_properties.get_description_foreach_http_status_code()
        for status_code, description in status_code_to_description_map.items():
            if status_code == 200:
                response_headers = self.arg_option_handler.extract_response_headers_from_arg_options(
                    url_method_properties.arg_options
                )
                return_value = self._build_return_value_wrapper(url_method_properties)
                result[str(status_code)] = Response(description=description, content=return_value, headers=response_headers)
            else:
                result[str(status_code)] = Response(description=description)

        return result

    def _build_return_value_wrapper(self, url_method_properties: MethodProperties) -> Optional[dict[str, MediaType]]:
        return_type = inspect.signature(url_method_properties.function).return_annotation

        if return_type is None or return_type == inspect.Signature.empty:
            return None

        return_properties: Optional[dict[str, Schema]] = None

        if return_type == ReturnValue or is_generic_type(return_type) and get_origin(return_type) == ReturnValue:
            # Dealing with the special case of ReturnValue[...]
            links_type = self.type_converter.get_openapi_type(dict[str, str])
            links_type.title = "Links"
            links_type.nullable = True

            warnings_type = self.type_converter.get_openapi_type(list[str])
            warnings_type.title = "Warnings"

            return_properties = {
                "links": links_type,
                "metadata": Schema(
                    title="Metadata",
                    nullable=True,
                    type="object",
                    properties={
                        "warnings": warnings_type,
                    },
                ),
            }

            type_args = get_args(return_type, evaluate=True)
            if not type_args or len(type_args) != 1:
                raise RuntimeError(
                    "ReturnValue definition should take one type Argument, e.g. ReturnValue[None].  "
                    f"Got this instead: {type_args}"
                )

            if not url_method_properties.envelope:
                raise RuntimeError("Methods returning a ReturnValue object should always have an envelope")

            if type_args[0] is not type(None):
                return_properties[url_method_properties.envelope_key] = self.type_converter.get_openapi_type(type_args[0])

        else:
            openapi_return_type = self.type_converter.get_openapi_type(return_type)
            if url_method_properties.envelope:
                return_properties = {url_method_properties.envelope_key: openapi_return_type}

        return {"application/json": MediaType(schema=Schema(type="object", properties=return_properties))}



"""
Copyright 2023 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import secrets
import socket

import asyncpg
import click
from asyncpg import PostgresError

import nacl.pwhash
from inmanta import config, data
from inmanta.const import MIN_PASSWORD_LENGTH
from inmanta.data.model import AuthMethod
from inmanta.protocol import auth
from inmanta.server import config as server_config


class ConnectionPoolException(Exception):
    """Exception that is raised if the connection to the connection pool fails"""


def generate_signing_config() -> str:
    hostname = socket.gethostname()
    return f"""[auth_jwt_default]
algorithm=HS256
sign=true
client_types=agent,compiler,api
key={secrets.token_urlsafe()}
expire=0
issuer=https://{hostname}:{server_config.server_bind_port.get()}/
audience=https://{hostname}:{server_config.server_bind_port.get()}/
"""


def validate_server_setup() -> None:
    """Validate the server configuration so that authentication is setup correctly."""
    config.Config.load_config(config_dir="/etc/inmanta/inmanta.d")

    value = click.confirm(
        "This command should be execute locally on the orchestrator you want to configure. Are you "
        "running this command locally?"
    )
    if value:
        # make sure auth is on
        if not server_config.server_enable_auth.get():
            raise click.ClickException(
                "Server authentication should be enabled before running the initial user setup. "
                "The option auth in the server section should be enabled."
            )

        click.echo(f"{'Server authentication: ' : <50}{click.style('enabled', fg='green')}")

        # make sure the method is set to database
        if server_config.server_auth_method.get() != "database":
            raise click.ClickException(
                "The server authentication method should be set to database to continue. Make sure auth_method in the server "
                "section is set to database"
            )

        click.echo(f"{'Server authentication method: ' : <50}{click.style('database', fg='green')}")

        # make sure there is auth config that supports signing tokens
        cfg = auth.AuthJWTConfig.get_sign_config()
        if cfg is None:
            click.echo("Error: No signing config available in the configuration.")
            click.echo("To use a new config, add the following to the configuration in /etc/inmanta/inmanta.d/auth.cfg:\n")
            click.echo(generate_signing_config())
            raise click.ClickException(
                "Make sure signing configuration is added to the config. See the documentation for details."
            )

        click.echo(f"{'Authentication signing config: ' : <50}{click.style('found', fg='green')}")

    else:
        raise click.ClickException("The user setup was aborted as it was not executed locally on the orchestrator")


async def get_connection_pool() -> asyncpg.pool.Pool:
    """
    connect to a connection pool. It is the responsibility of the caller to close the return connection pool
    """
    try:
        database_host = server_config.db_host.get()
        database_port = server_config.db_port.get()

        database_username = server_config.db_username.get()
        database_password = server_config.db_password.get()
        connection_pool_min_size = server_config.server_db_connection_pool_min_size.get()
        connection_pool_max_size = server_config.server_db_connection_pool_max_size.get()
        connection_timeout = server_config.db_connection_timeout.get()
        return await data.connect(
            database_host,
            database_port,
            server_config.db_name.get(),
            database_username,
            database_password,
            create_db_schema=False,
            connection_pool_min_size=connection_pool_min_size,
            connection_pool_max_size=connection_pool_max_size,
            connection_timeout=connection_timeout,
        )
    except Exception:
        raise ConnectionPoolException()


async def do_user_setup() -> None:
    """Perform the user setup that requires the database interaction"""
    connection = None
    try:
        click.echo(
            f"{'Trying to connect to DB:': <50}"
            f"{('%s (%s:%s)' % (server_config.db_name.get(), server_config.db_host.get(), server_config.db_port.get()))}"
        )
        connection = await get_connection_pool()
        click.echo(f"{'Connection to database' : <50}{click.style('success', fg='green')}")
        users = await data.User.get_list()

        if len(users):
            raise click.ClickException("There are already users in the database")

        username = click.prompt("What username do you want to use?", default="admin")
        if not username:
            raise click.ClickException("the username cannot be an empty string")

        password = click.prompt("What password do you want to use?", hide_input=True)
        if not password or len(password) < MIN_PASSWORD_LENGTH:
            raise click.ClickException("the password should be at least 8 characters long")

        pw_hash = nacl.pwhash.str(password.encode())

        # insert the user
        user = data.User(
            username=username,
            password_hash=pw_hash.decode(),
            auth_method=AuthMethod.database,
        )
        await user.insert()

        click.echo(f"{'User %s: ' % username : <50}{click.style('created', fg='green')}")
    except ConnectionPoolException as e:
        click.echo(f"{f'Connection to database {server_config.db_host.get()}' : <50}" f"{click.style('failed', fg='red')}")
        raise e
    except PostgresError:
        raise click.ClickException(
            "The version of the database is out of date: start the server to upgrade the database "
            "schema to the lastest version."
        )

    finally:
        if connection is not None:
            await data.disconnect()

    click.echo("Make sure to (re)start the orchestrator to activate all changes.")


@click.command(help="Do the initial user setup. This command should be executed on the orchestrator and not remotely.")
def cmd() -> None:
    # validate the setup so that we can setup a new user
    validate_server_setup()
    # check if there are already users
    asyncio.run(do_user_setup())


def main() -> None:
    cmd()


if __name__ == "__main__":
    main()



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    schema = """
     UPDATE public.compile SET environment_variables = '{}' WHERE environment_variables is null;

     ALTER TABLE public.compile
        ADD COLUMN mergeable_environment_variables jsonb DEFAULT '{}' NOT NULL,
        ADD COLUMN used_environment_variables jsonb,
        ALTER COLUMN environment_variables SET NOT NULL;

     ALTER TABLE public.compile RENAME COLUMN environment_variables TO requested_environment_variables;

     UPDATE public.compile SET used_environment_variables = requested_environment_variables WHERE completed is not null;
     """

    await connection.execute(schema)



"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import datetime
import json
import os
import typing
import urllib
import uuid
from collections import abc
from collections.abc import Sequence
from enum import Enum, StrEnum
from typing import ClassVar, Mapping, Optional, Self, Union, assert_never

import pydantic.schema
from pydantic import ConfigDict, Field, computed_field, field_validator, model_validator

import inmanta
import inmanta.ast.export as ast_export
import pydantic_core.core_schema
from inmanta import const, data, protocol, resources
from inmanta.stable_api import stable_api
from inmanta.types import ArgumentTypes
from inmanta.types import BaseModel as BaseModel  # Keep in place for backwards compat with <=ISO8
from inmanta.types import JsonType
from inmanta.types import ResourceIdStr as ResourceIdStr  # Keep in place for backwards compat with <=ISO8
from inmanta.types import ResourceType as ResourceType  # Keep in place for backwards compat with <=ISO8
from inmanta.types import ResourceVersionIdStr as ResourceVersionIdStr  # Keep in place for backwards compat with <=ISO8
from inmanta.types import SimpleTypes


class ExtensionStatus(BaseModel):
    """
    Status response for extensions loaded in the server
    """

    name: str
    version: str
    package: str


class ReportedStatus(StrEnum):
    OK = "OK"
    Warning = "Warning"
    Error = "Error"

    def __gt__(self, other: str) -> bool:
        # Determines the order of severity of the reported status
        order: list[str] = [ReportedStatus.OK, ReportedStatus.Warning, ReportedStatus.Error]
        if self not in order or other not in order:
            raise ValueError
        return order.index(self) > order.index(other)


class SliceStatus(BaseModel):
    """
    Status response for slices loaded in the server
    """

    name: str
    status: Mapping[str, ArgumentTypes | Mapping[str, ArgumentTypes]]
    reported_status: ReportedStatus
    message: str | None = None


class FeatureStatus(BaseModel):
    """
    Status of the feature
    """

    slice: str
    name: str
    value: Optional[object] = None


class StatusResponse(BaseModel):
    """
    Response for the status method call

    :param product: The name of the product.
    :param edition: The edition of the product.
    :param version: The version of the product.
    :param license: The license used by the product.
    :param extensions: The status of the extensions of the server
    :param slices: The status of the slices of the server.
    :param features: The status of the features offered by the slices of the server.
    :param status: The overall status of the server
    :param python_version: The python version used by the server.
    :param postgresql_version: The postgresql version used by the database slice
        None if it is not initialized or an error occurred with the database slice.
    """

    product: str
    edition: str
    version: str
    license: Union[str, dict[str, SimpleTypes]]
    extensions: list[ExtensionStatus]
    slices: list[SliceStatus]
    features: list[FeatureStatus]
    status: ReportedStatus
    python_version: str
    postgresql_version: str | None


@stable_api
class CompileData(BaseModel):
    """
    Top level structure of compiler data to be exported.
    """

    errors: list[ast_export.Error]
    """
        All errors occurred while trying to compile.
    """


class CompileRunBase(BaseModel):
    """
    :param requested_environment_variables: environment variables requested to be passed to the compiler
    :param mergeable_environment_variables: environment variables to be passed to the compiler.
            These env vars can be compacted over multiple compiles.
            If multiple values are compacted, they will be joined using spaces.
    :param environment_variables: environment variables passed to the compiler
    :param links: An object that contains relevant links to this compile.
        It is a dictionary where the key is something that identifies one or more links
        and the value is a list of urls. i.e. {"instances": ["link-1',"link-2"], "compiles": ["link-3"]}
    """

    id: uuid.UUID
    remote_id: Optional[uuid.UUID] = None
    environment: uuid.UUID
    requested: Optional[datetime.datetime] = None
    started: Optional[datetime.datetime] = None

    do_export: bool
    force_update: bool
    metadata: JsonType
    mergeable_environment_variables: dict[str, str]
    requested_environment_variables: dict[str, str]
    environment_variables: dict[str, str]

    partial: bool
    removed_resource_sets: list[str]

    exporter_plugin: Optional[str] = None

    notify_failed_compile: Optional[bool] = None
    failed_compile_message: Optional[str] = None
    links: dict[str, list[str]] = {}

    @pydantic.field_validator("environment_variables", mode="before")
    @classmethod
    def validate_environment_variables(cls, v: typing.Any, info: pydantic_core.core_schema.ValidationInfo) -> typing.Any:
        """
        Default the environment_variables to requested_environment_variables + mergeable_environment_variables

        This relies on the fact that fields are validated in the order they are declared!
        """
        if v is None:
            out = {}
            out.update(info.data["requested_environment_variables"])
            out.update(info.data["mergeable_environment_variables"])
            return out
        else:
            return v


class CompileRun(CompileRunBase):
    compile_data: Optional[CompileData] = None


class CompileReport(CompileRunBase):
    completed: Optional[datetime.datetime] = None
    success: Optional[bool] = None
    version: Optional[int] = None


class CompileRunReport(BaseModel):
    id: uuid.UUID
    started: datetime.datetime
    completed: Optional[datetime.datetime] = None
    command: str
    name: str
    errstream: str
    outstream: str
    returncode: Optional[int] = None


class CompileDetails(CompileReport):
    compile_data: Optional[CompileData] = None
    reports: Optional[list[CompileRunReport]] = None


class AttributeStateChange(BaseModel):
    """
    Changes in the attribute
    """

    current: Optional[object] = None
    desired: Optional[object] = None

    @field_validator("current", "desired")
    @classmethod
    def check_serializable(cls, v: Optional[object]) -> Optional[object]:
        """
        Verify whether the value is serializable (https://github.com/inmanta/inmanta-core/issues/3470)
        """
        try:
            protocol.common.json_encode(v)
        except TypeError:
            if inmanta.RUNNING_TESTS:
                # Fail the test when the value is not serializable
                raise Exception(f"Failed to serialize attribute {v}")
            else:
                # In production, try to cast the non-serializable value to str to prevent the handler from failing.
                return str(v)
        return v

    def __getstate__(self) -> str:
        # make pickle use json to keep from leaking stuff
        # Will make the objects into json-like things
        # This method exists only to keep IPC light compatible with the json based RPC
        return protocol.common.json_encode(self)

    def __setstate__(self, state: str) -> None:
        # This method exists only to keep IPC light compatible with the json based RPC
        self.__dict__.update(json.loads(state))


EnvSettingType = Union[bool, int, float, str, dict[str, Union[str, int, bool]]]


class Environment(BaseModel):
    """
    An inmanta environment.

    :note: repo_url and repo_branch will be moved to the settings.
    """

    id: uuid.UUID
    name: str
    project_id: uuid.UUID
    repo_url: str
    repo_branch: str
    settings: dict[str, EnvSettingType]
    halted: bool
    is_marked_for_deletion: bool = False
    description: Optional[str] = None
    icon: Optional[str] = None


class Project(BaseModel):
    """
    An inmanta project.
    """

    id: uuid.UUID
    name: str
    environments: list[Environment]


class EnvironmentSetting(BaseModel):
    """A class to define a new environment setting.

    :param name: The name of the setting.
    :param type: The type of the value. This type is mainly used for documentation purpose.
    :param default: An optional default value for this setting. When a default is set and the
                    is requested from the database, it will return the default value and also store
                    the default value in the database.
    :param doc: The documentation/help string for this setting
    :param recompile: Trigger a recompile of the model when a setting is updated?
    :param update_model: Update the configuration model (git pull on project and repos)
    :param agent_restart: Restart autostarted agents when this settings is updated.
    :param allowed_values: list of possible values (if type is enum)
    """

    name: str
    type: str
    default: EnvSettingType
    doc: str
    recompile: bool
    update_model: bool
    agent_restart: bool
    allowed_values: Optional[list[EnvSettingType]] = None


class ProtectedBy(str, Enum):
    """
    An enum that indicates the reason why an environment setting can be protected.
    """

    # The environment setting is managed using the environment_settings property of the project.yml file.
    project_yml = "project.yml"

    def get_detailed_description(self) -> str:
        """
        Return a string that explains in detail why the environment setting is protected.
        """
        match self:
            case ProtectedBy.project_yml:
                return "Setting is managed by the project.yml file of the Inmanta project."
            case _ as unreachable:
                assert_never(unreachable)

    @classmethod
    def _missing_(cls: type[Self], value: object) -> Optional[Self]:
        """
        This is a workaround for the issue where the protocol layer inconsistently handles enums.
        Enums are serialized using their name, but deserialized using their value. This method makes
        sure that we can deserialize enums using their name.
        """
        return next((p for p in cls if p.name == value), None) if isinstance(value, str) else None


class EnvironmentSettingDefinitionAPI(EnvironmentSetting):
    """
    The definition of an environment setting as served out over the API.
    """

    protected: bool = False
    protected_by: ProtectedBy | None = None


class EnvironmentSettingDetails(BaseModel):
    """
    A class that stores the value and other metadata about an environment setting.

    :param value: The value of the environment setting.
    :param protected: True iff the environment setting cannot be updated using the normal
                      endpoints to update environment settings.
    :param protected_by: This field indicates the reason why the environment setting is protected.
                         This field is set to None if the environment setting is not protected.
    """

    value: EnvSettingType
    protected: bool = False
    protected_by: ProtectedBy | None = None


class EnvironmentSettingsReponse(BaseModel):

    settings: dict[str, EnvSettingType]
    definition: dict[str, EnvironmentSettingDefinitionAPI]


class ModelMetadata(BaseModel):
    """Model metadata"""

    inmanta_compile_state: const.Compilestate = Field(default=const.Compilestate.success, alias="inmanta:compile:state")
    message: str
    type: str
    extra_data: Optional[JsonType] = None


class ResourceMinimal(BaseModel):
    """
    Represents a resource object as it comes in over the API. Provides strictly required validation only.
    """

    id: ResourceVersionIdStr

    @field_validator("id")
    @classmethod
    def id_is_resource_version_id(cls, v):
        if resources.Id.is_resource_version_id(v):
            return v
        raise ValueError(f"id {v} is not of type ResourceVersionIdStr")

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="allow")


class Resource(BaseModel):
    environment: uuid.UUID
    model: int
    resource_id: ResourceIdStr
    resource_type: ResourceType
    resource_version_id: ResourceVersionIdStr
    resource_id_value: str
    agent: str
    attributes: JsonType
    status: const.ResourceState
    resource_set: Optional[str] = None


class ResourceAction(BaseModel):
    environment: uuid.UUID
    version: int
    resource_version_ids: list[ResourceVersionIdStr]
    action_id: uuid.UUID
    action: const.ResourceAction
    started: datetime.datetime
    finished: Optional[datetime.datetime] = None
    messages: Optional[list[JsonType]] = None
    status: Optional[const.ResourceState] = None
    changes: Optional[JsonType] = None
    change: Optional[const.Change] = None
    send_event: Optional[bool] = None  # Deprecated field


class ResourceDeploySummary(BaseModel):
    """
    :param total: The total number of resources
    :param by_state: The number of resources by state in the latest released version
    """

    total: int
    by_state: dict[str, int]

    @classmethod
    def create_from_db_result(cls, summary_by_state: dict[str, int]) -> "ResourceDeploySummary":
        full_summary_by_state = cls._ensure_summary_has_all_states(summary_by_state)
        total = cls._count_all_resources(full_summary_by_state)
        return ResourceDeploySummary(by_state=full_summary_by_state, total=total)

    @classmethod
    def _ensure_summary_has_all_states(cls, summary_by_state: dict[str, int]) -> dict[str, int]:
        full_summary = summary_by_state.copy()
        for state in const.ResourceState:
            if state not in summary_by_state.keys() and state != const.ResourceState.dry:
                full_summary[state] = 0
        return full_summary

    @classmethod
    def _count_all_resources(cls, summary_by_state: dict[str, int]) -> int:
        return sum(resource_count for resource_count in summary_by_state.values())


class LogLine(BaseModel):
    # Override the setting from the BaseModel class as such that the level field is
    # serialized using the name of the enum instead of its value. This is required
    # to make sure that data sent to the API endpoints resource_action_update
    # and send_deploy_done are serialized consistently using the name of the enum.
    model_config: ClassVar[ConfigDict] = ConfigDict(use_enum_values=False)

    level: const.LogLevel
    msg: str
    args: list[Optional[ArgumentTypes]] = []
    kwargs: JsonType = {}
    timestamp: datetime.datetime

    @field_validator("level", mode="before")
    @classmethod
    def validate_log_level(cls, value: object) -> const.LogLevel:
        """
        Validate the log level using the LogLevel enum. Pydantic's default validation does not suffice because of the
        custom value aliasing behavior built on top of LogLevel to allow passing ints to the constructor.
        """
        try:
            return const.LogLevel(value)
        except ValueError:
            # error message as close to pydantic's as possible but add in the int aliases
            name_value_pairs: abc.Iterator[tuple[str, int]] = ((level.value, level.to_int) for level in const.LogLevel)
            valid_input_descriptions: list[str] = [f"'{name}' | {num_value}" for name, num_value in name_value_pairs]
            raise ValueError(
                "Input should be %s" % " or ".join((", ".join(valid_input_descriptions[:-1]), valid_input_descriptions[-1]))
            )


class ResourceIdDetails(BaseModel):
    resource_type: ResourceType
    agent: str
    attribute: str
    resource_id_value: str


class ReleasedResourceState(StrEnum):
    # Copied over from const.ResourceState
    unavailable = "unavailable"  # This state is set by the agent when no handler is available for the resource
    skipped = "skipped"  #
    dry = "dry"
    deployed = "deployed"
    failed = "failed"
    deploying = "deploying"
    available = "available"
    cancelled = "cancelled"  # When a new version is pushed, in progress deploys are cancelled
    undefined = "undefined"  # The state of this resource is unknown at this moment in the orchestration process
    skipped_for_undefined = "skipped_for_undefined"  # This resource depends on an undefined resource
    orphaned = "orphaned"


class VersionedResource(BaseModel):
    resource_id: ResourceIdStr
    resource_version_id: ResourceVersionIdStr
    id_details: ResourceIdDetails
    requires: list[ResourceVersionIdStr]

    @property
    def all_fields(self) -> dict[str, object]:
        return {**self.dict(), **self.id_details.dict()}


class LatestReleasedResource(VersionedResource):
    status: ReleasedResourceState


class PagingBoundaries:
    """
    Represents the lower and upper bounds that should be used for the next and previous pages
    when listing domain entities.

    The largest / smallest value of the current page represents respectively the min / max boundary value (exclusive) for the
    neighbouring pages. Which represents next and which prev depends on sorting order (ASC or DESC).
    So, while the names "start" and "end" might seem to indicate "left" and "right" of the page, they actually mean "highest"
    and "lowest".

    Let's show this in an example: a user requests the following:
     - all Resources with name > foo
     - ASCENDING order
     - Page size = 10

    The equivalent RequestPagingBoundary will be as follows:
        ```
        RequestPagingBoundary:
            start = foo
            end = None
        ```

    The fetched data will be: [foo1 ... foo10]

    But the Pagingboundary will be constructed this way:
        ```
        Pagingboundary:
            end = foo1
            start = foo10 # Reversed because these are meant to map to like-named fields on neighbouring RequestedPagingBoundary
        ```

    :param start: largest value of current page for the primary sort column.
    :param end: smallest value of current page for the primary sort column.
    :param first_id: largest value of current page for the secondary sort column, if there is one.
    :param last_id: smallest value of current page for the secondary sort column, if there is one.
    """

    def __init__(
        self,
        start: Optional["inmanta.data.PRIMITIVE_SQL_TYPES"],  # Can be none if user selected field is nullable
        end: Optional["inmanta.data.PRIMITIVE_SQL_TYPES"],  # Can be none if user selected field is nullable
        first_id: Optional["inmanta.data.PRIMITIVE_SQL_TYPES"],  # Can be none if single keyed
        last_id: Optional["inmanta.data.PRIMITIVE_SQL_TYPES"],  # Can be none if single keyed
    ) -> None:
        self.start = start
        self.end = end
        self.first_id = first_id
        self.last_id = last_id


class ResourceDetails(BaseModel):
    """The details of a resource
    :param resource_id: The id of the resource
    :param resource_type: The type of the resource
    :param agent: The agent associated with this resource
    :param id_attribute: The name of the identifying attribute of the resource
    :param id_attribute_value: The value of the identifying attribute of the resource
    :param attributes: The attributes of the resource
    """

    resource_id: ResourceIdStr
    resource_type: ResourceType
    agent: str
    id_attribute: str
    id_attribute_value: str
    attributes: JsonType


class VersionedResourceDetails(ResourceDetails):
    """The details of a resource version
    :param resource_version_id: The id of the resource
    :param version: The version of the resource
    """

    resource_version_id: ResourceVersionIdStr
    version: int

    @model_validator(mode="after")
    def ensure_version_field_set_in_attributes(self) -> Self:
        # Due to a bug, the version field has always been present in the attributes dictionary.
        # This bug has been fixed in the database. For backwards compatibility reason we here make sure that the
        # version field is present in the attributes dictionary served out via the API.
        if "version" not in self.attributes:
            self.attributes["version"] = self.version
        return self


class ReleasedResourceDetails(ResourceDetails):
    """The details of a released resource
    :param last_deploy: The value of the last_deploy on the latest released version of the resource
    :param first_generated_time: The first time this resource was generated
    :param first_generated_version: The first model version this resource was in
    :param status: The current status of the resource
    :param requires_status: The id and status of the resources this resource requires
    """

    last_deploy: Optional[datetime.datetime] = None
    first_generated_time: datetime.datetime
    first_generated_version: int
    status: ReleasedResourceState
    requires_status: dict[ResourceIdStr, ReleasedResourceState]


class ResourceHistory(BaseModel):
    resource_id: ResourceIdStr
    date: datetime.datetime
    attributes: JsonType
    attribute_hash: str
    requires: list[ResourceIdStr]


class ResourceLog(LogLine):
    action_id: uuid.UUID
    action: const.ResourceAction


class ResourceDiffStatus(str, Enum):
    added = "added"
    modified = "modified"
    deleted = "deleted"
    unmodified = "unmodified"
    agent_down = "agent_down"
    undefined = "undefined"
    skipped_for_undefined = "skipped_for_undefined"


class AttributeDiff(BaseModel):
    """
    :param from_value: The value of the attribute in the earlier version
    :param to_value: The value of the attribute in the later version
    :param from_value_compare: A stringified, diff-friendly form of the 'from_value' field
    :param to_value_compare: A stringified, diff-friendly form of the 'to_value' field
    """

    from_value: Optional[object] = None
    to_value: Optional[object] = None
    from_value_compare: str
    to_value_compare: str


class ResourceDiff(BaseModel):
    """
    :param resource_id: The id of the resource the diff is about (without version)
    :param attributes: The diff between the attributes of two versions of the resource
    :param status: The kind of diff between the versions of the resource
    """

    resource_id: ResourceIdStr
    attributes: dict[str, AttributeDiff]
    status: ResourceDiffStatus


class Parameter(BaseModel):
    id: uuid.UUID
    name: str
    value: str
    environment: uuid.UUID
    source: str
    updated: Optional[datetime.datetime] = None
    metadata: Optional[JsonType] = None


class Fact(Parameter):
    resource_id: ResourceIdStr
    expires: bool = True


class Agent(BaseModel):
    """
    :param environment: Id of the agent's environment
    :param name: The name of the agent
    :param last_failover: The time of the last failover
    :param paused: Whether the agent is paused or not
    :param unpause_on_resume: Whether the agent should be unpaused when the environment is resumed
    :param status: The current status of the agent
    :param process_id: The id of the agent process that belongs to this agent, if there is one
    :param process_name: The name of the agent process that belongs to this agent, if there is one
    """

    environment: uuid.UUID
    name: str
    last_failover: Optional[datetime.datetime] = None
    paused: bool
    process_id: Optional[uuid.UUID] = None
    process_name: Optional[str] = None
    unpause_on_resume: Optional[bool] = None
    status: const.AgentStatus


class AgentProcess(BaseModel):
    sid: uuid.UUID
    hostname: str
    environment: uuid.UUID
    first_seen: Optional[datetime.datetime] = None
    last_seen: Optional[datetime.datetime] = None
    expired: Optional[datetime.datetime] = None
    state: Optional[dict[str, Union[dict[str, list[str]], dict[str, str], dict[str, float], str]]] = None


class DesiredStateLabel(BaseModel):
    name: str
    message: str


class DesiredStateVersion(BaseModel):
    """
    :param released: has this desired state version been released?
    """

    version: int
    date: datetime.datetime
    total: int
    labels: list[DesiredStateLabel]
    status: const.DesiredStateVersionStatus
    released: bool


class PromoteTriggerMethod(StrEnum):
    # partly copies from const.AgentTriggerMethod
    push_incremental_deploy = "push_incremental_deploy"
    push_full_deploy = "push_full_deploy"
    no_push = "no_push"


class DryRun(BaseModel):
    id: uuid.UUID
    environment: uuid.UUID
    model: int
    date: Optional[datetime.datetime] = None
    total: int = 0
    todo: int = 0


class DryRunReport(BaseModel):
    summary: DryRun
    diff: list[ResourceDiff]


class Notification(BaseModel):
    """
    :param id: The id of this notification
    :param environment: The environment this notification belongs to
    :param created: The date the notification was created at
    :param title: The title of the notification
    :param message: The actual text of the notification
    :param severity: The severity of the notification
    :param uri: A link to an api endpoint of the server, that is relevant to the message,
                and can be used to get further information about the problem.
                For example a compile related problem should have the uri: `/api/v2/compilereport/<compile_id>`
    :param compile_id: The id of the compile that is associated with this notification.
    :param read: Whether the notification was read or not
    :param cleared: Whether the notification was cleared or not
    """

    id: uuid.UUID
    environment: uuid.UUID
    created: datetime.datetime
    title: str
    message: str
    severity: const.NotificationSeverity
    uri: Optional[str] = None
    compile_id: uuid.UUID | None = None
    read: bool
    cleared: bool


class Source(BaseModel):
    """Model for source code"""

    hash: str
    is_byte_code: bool
    module_name: str
    requirements: list[str]


class EnvironmentMetricsResult(BaseModel):
    """
    A container for metrics as returned by the /metrics endpoint.

    :param start: The starting of the aggregation interval.
    :param end: The end of the aggregation interval.
    :param timestamps: The timestamps that belongs to the aggregated metrics present in the `metrics` dictionary.
    :param metrics: A dictionary that maps the name of a metric to a list of aggregated datapoints. For metrics that are not
                    grouped on a specific property, this list only contains the values of the metrics. For metrics that
                    are grouped by a specific property, this list contains a dictionary where the key is the grouping
                    attribute and the value is the value of the metric. The value is None when no data is available
                    for that specific time window.
    """

    start: datetime.datetime
    end: datetime.datetime
    timestamps: list[datetime.datetime]
    metrics: dict[str, list[Optional[Union[float, dict[str, float]]]]]


class AuthMethod(str, Enum):
    database = "database"
    oidc = "oidc"


class User(BaseModel):
    """A user"""

    username: str
    auth_method: AuthMethod


class CurrentUser(BaseModel):
    """Information about the current logged in user"""

    username: str


class LoginReturn(BaseModel):
    """
    Login information

    :param token: A token representing the user's authentication session
    :param user: The user object for which the token was created
    """

    token: str
    user: User


def _check_resource_id_str(v: str) -> ResourceIdStr:
    if resources.Id.is_resource_id(v):
        return ResourceIdStr(v)
    raise ValueError("Invalid id for resource %s" % v)


ResourceId: typing.TypeAlias = typing.Annotated[ResourceIdStr, pydantic.AfterValidator(_check_resource_id_str)]


class DiscoveredResourceABC(BaseModel):
    """
    :param discovered_resource_id: The name of the resource
    :param values: The actual resource
    :param managed_resource_uri: URI of the resource with the same ID that is already
        managed by the orchestrator e.g. "/api/v2/resource/<rid>". Or None if the resource is not managed.
    :param discovery_resource_id: Resource id of the (managed) discovery resource that reported this
        discovered resource.
    """

    discovered_resource_id: ResourceId
    values: dict[str, object]
    managed_resource_uri: Optional[str] = None

    discovery_resource_id: Optional[ResourceId]

    @computed_field  # type: ignore[misc]
    @property
    def discovery_resource_uri(self) -> str | None:
        if self.discovery_resource_id is None:
            return None
        return f"/api/v2/resource/{urllib.parse.quote(self.discovery_resource_id, safe='')}"


class DiscoveredResource(DiscoveredResourceABC):
    """
    Discovered resource for API returns. Contains additional (redundant) metadata to improve user experience.
    """

    resource_type: ResourceType
    agent: str
    resource_id_value: str


class LinkedDiscoveredResource(DiscoveredResourceABC):
    """
    DiscoveredResource linked to the discovery resource that discovered it.

    :param discovery_resource_id: Resource id of the (managed) discovery resource that reported this
           discovered resource.
    """

    # This class is used as API input. Its behaviour can be directly incorporated into the DiscoveredResourceABC parent class
    # when providing the id of the discovery resource is mandatory for all discovered resource. Ticket link:
    # https://github.com/inmanta/inmanta-core/issues/8004

    discovery_resource_id: ResourceId

    def to_dao(self, env: uuid.UUID) -> "data.DiscoveredResource":
        parsed_id: resources.Id = resources.Id.parse_id(self.discovered_resource_id)
        return data.DiscoveredResource(
            discovered_resource_id=self.discovered_resource_id,
            resource_type=parsed_id.entity_type,
            agent=parsed_id.agent_name,
            resource_id_value=parsed_id.attribute_value,
            values=self.values,
            discovered_at=datetime.datetime.now(),
            environment=env,
            discovery_resource_id=self.discovery_resource_id,
        )


def hyphenize(field: str) -> str:
    """Alias generator to convert python names (with `_`) to config file name (with `-`)"""
    return field.replace("_", "-")


@stable_api
# This is part of both the config file schema and the api schema
class PipConfig(BaseModel):
    """
    Base class to represent pip config internally

    :param index_url: one pip index url for this project.
    :param extra_index_url:  additional pip index urls for this project. This is generally only
        recommended if all configured indexes are under full control of the end user to protect against dependency
        confusion attacks. See the `pip install documentation <https://pip.pypa.io/en/stable/cli/pip_install/>`_ and
        `PEP 708 (draft) <https://peps.python.org/pep-0708/>`_ for more information.
    :param pre:  allow pre-releases when installing Python packages, i.e. pip --pre.
        Defaults to None.
        When None and pip.use-system-config=true we follow the system config.
        When None and pip.use-system-config=false, we don't allow pre-releases.
    :param use_system_config: defaults to false.
        When true, sets the pip index url, extra index urls and pre according to the respective settings outlined above
        but otherwise respect any pip environment variables and/or config in the pip config file,
        including any extra-index-urls.

        If no indexes are configured in pip.index-url/pip.extra-index-url
        with this option enabled means to fall back to pip's default behavior:
        use the pip index url from the environment, the config file, or PyPi, in that order.

        For development, it is recommended to set this option to false, both for portability
        (and related compatibility with tools like pytest-inmanta-lsm) and for security
        (dependency confusion attacks could affect users that aren't aware that inmanta installs Python packages).
    """

    # Config needs to be in the top-level object, because is also affect serialization/deserialization
    model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        # use alias generator have `-` in names
        alias_generator=hyphenize,
        # allow use of aliases
        populate_by_name=True,
        extra="ignore",
    )

    index_url: Optional[str] = None
    # Singular to be consistent with pip itself
    extra_index_url: Sequence[str] = []
    pre: Optional[bool] = None
    use_system_config: bool = False

    def has_source(self) -> bool:
        """Can this config get packages from anywhere?"""
        return bool(self.index_url) or self.use_system_config

    def get_index_args(self) -> list[str]:
        """
        Returns the index-related arguments that should be used to run a pip command
        with this pip config.
        """
        index_args: list[str] = []
        if self.index_url:
            index_args.append("--index-url")
            index_args.append(self.index_url)
        elif not self.use_system_config:
            # If the config doesn't set index url
            # and we are not using system config,
            # then we need to disable the index.
            # This can only happen if paths is also set.
            index_args.append("--no-index")
        for extra_index_url in self.extra_index_url:
            index_args.append("--extra-index-url")
            index_args.append(extra_index_url)
        return index_args

    def get_environment_variables(self) -> dict[str, str]:
        """
        Returns the environment variables that should be used to run a pip command
        with this pip config.
        """
        sub_env = os.environ.copy()
        if not self.use_system_config:
            # If we don't use system config, unset env vars
            for key in ("PIP_EXTRA_INDEX_URL", "PIP_INDEX_URL", "PIP_PRE", "PIP_NO_INDEX"):
                sub_env.pop(key, None)

            # setting this env_var to os.devnull disables the loading of all pip configuration file
            sub_env["PIP_CONFIG_FILE"] = os.devnull
        if self.pre is not None:
            # Make sure that IF pip pre is set, we enforce it
            # The `--pre` option can only enable it
            # The env var can both enable and disable
            sub_env["PIP_PRE"] = str(self.pre)
        return sub_env


LEGACY_PIP_DEFAULT = PipConfig(use_system_config=True)


class Discrepancy(BaseModel):
    """
    Records a discrepancy between the state as persisted in the database and
    the in-memory state in the scheduler. Either model-wide when no
    resource id is specified (e.g. when model versions are mismatched)
    or for a specific resource.

    :param rid: If set, this discrepancy is specific to this resource.
        If left unset, this discrepancy is not specific to any particular resource.
    :param field: If set, specifies on which field this discrepancy was detected.
        If left unset, and a rid is specified, the discrepancy was detected on the
        resource level i.e. it is missing from either the db or the scheduler.
    :param expected: User-facing message denoting the expected state (i.e. as persisted
        in the DB).
    :param actual: User-facing message denoting the actual state (i.e. in-memory state
        in the scheduler).

    """

    rid: ResourceIdStr | None
    field: str | None
    expected: str
    actual: str


class SchedulerStatusReport(BaseModel):
    """
    Status report for the scheduler self-check

    :param scheduler_state: In-memory representation of the resources in the scheduler
    :param db_state: Desired state of the resources as persisted in the database
    :param discrepancies: Discrepancies between the in-memory representation of the resources
        and their state in the database.
    """

    # Can't type properly because of current module structure
    scheduler_state: Mapping[ResourceIdStr, object]  # "True" type is deploy.state.ResourceState
    db_state: Mapping[ResourceIdStr, object]  # "True" type is deploy.state.ResourceIntent
    resource_states: Mapping[ResourceIdStr, const.ResourceState]
    discrepancies: list[Discrepancy] | dict[ResourceIdStr, list[Discrepancy]]


class DataBaseReport(BaseModel):
    """
    :param max_pool: maximal pool size
    :param free_pool: number of connections not in use in the pool
    :param open_connections: number of connections currently open
    :param free_connections: number of connections currently open and not in use
    :param pool_exhaustion_time: nr of seconds since start we observed the pool to be exhausted
    """

    connected: bool
    database: str
    host: str
    max_pool: int
    free_pool: int
    open_connections: int
    free_connections: int
    pool_exhaustion_time: float

    def __add__(self, other: "DataBaseReport") -> "DataBaseReport":
        if not isinstance(other, DataBaseReport):
            return NotImplemented
        if other.database != self.database:
            return NotImplemented
        if other.host != self.host:
            return NotImplemented
        return DataBaseReport(
            connected=self.connected and other.connected,
            database=self.database,
            host=self.host,
            max_pool=self.max_pool + other.max_pool,
            free_pool=self.free_pool + other.free_pool,
            open_connections=self.open_connections + other.open_connections,
            free_connections=self.free_connections + other.free_connections,
            pool_exhaustion_time=self.pool_exhaustion_time + other.pool_exhaustion_time,
        )


class GetSourceCodeResponse(BaseModel):
    sources: list[Source]
    project_constraints: str | None



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import builtins
import copy
import dataclasses
import functools
import numbers
from collections import defaultdict, deque
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Optional

from inmanta import references
from inmanta.ast import (
    DuplicateException,
    Locatable,
    LocatableString,
    Location,
    MultiUnsetException,
    Named,
    Namespace,
    NotFoundException,
    RuntimeException,
    TypingException,
    UnexpectedReference,
    UnsetException,
)
from inmanta.execute.proxy import DynamicProxy, ProxyContext
from inmanta.execute.util import AnyType, NoneValue, Unknown
from inmanta.stable_api import stable_api

if TYPE_CHECKING:
    from inmanta.ast.statements import ExpressionStatement


@stable_api
class Type(Locatable):
    """
    This class is the abstract base class for all types in the Inmanta :term:`DSL` that represent basic data. These are
    types that are not relations. Instances of subclasses represent a type in the Inmanta language.
    """

    def validate(self, value: Optional[object]) -> bool:
        """
        Validate the given value to check if it satisfies the constraints associated with this type. Returns true iff
        validation succeeds, otherwise raises a :py:class:`inmanta.ast.RuntimeException`.

        In advanced cases where this class has a custom ``to_python()``, translation-specific validation should be deferred
        to that stage. Translation-specific means that the value is definitely of this DSL type, but it can not be converted to
        the Python domain.
        """
        # Special DSL values like references require an explicit annotation so we don't leak them where they aren't expected.
        if isinstance(value, references.Reference):
            raise UnexpectedReference(
                reference=value,
                # keep message generic, since this method is used for many types' super() call.
                message=(
                    f"References are not allowed for values of type `{self.type_string_internal()}`. To work with references,"
                    " explicitly declare support with a `... | Reference[...]` annotation."
                ),
            )
        return True

    def type_string(self) -> Optional[str]:
        """
        Returns the type string as expressed in the Inmanta :term:`DSL`, if this type can be expressed in the :term:`DSL`.
        Otherwise returns None.
        """
        return None

    def type_string_internal(self) -> str:
        """
        Returns the internal string representation of the instance of the type. This is used by __str__
        when type_string() returns None.
        Use this method only when you explicitly need the internal string representation of the type.
        """
        return "Type"

    def __str__(self) -> str:
        """
        Returns the string representation of the type, to be used for informative reporting as in error messages.
        When a structured representation of the inmanta type is required, type_string() should be used instead.
        """
        type_string: Optional[str] = self.type_string()
        return type_string if type_string is not None else self.type_string_internal()

    def normalize(self) -> None:
        pass

    def is_attribute_type(self) -> bool:
        """
        Returns true iff this type is valid in the model as an attribute type
        """
        return False

    def is_entity(self) -> bool:
        """
        Returns true only for Entity
        """
        # Introduced to prevent import loops on isinstance checks
        return False

    def get_base_type(self) -> "Type":
        """
        Returns the base type for this type, i.e. the plain type without modifiers such as expressed by
        ``[]`` and ``?`` in the :term:`DSL` and ``Reference`` in the plugin domain.
        """
        return self

    def get_no_reference(self) -> "Type":
        """
        Returns the same type, but with all references removed
        """
        return self

    def supports_references(self) -> bool:
        """
        Returns True iff this type accepts any sort of reference value.
        """
        return self.get_no_reference() is not self

    def with_base_type(self, base_type: "Type") -> "Type":
        """
        Returns the type formed by replacing this type's base type with the supplied type.
        """
        return base_type

    def corresponds_to(self, type: "Type") -> bool:
        """
        Determine if the given 'type' is a good approximation to `self` given that `type` is derived from a python type

        Intended specifically for type correspondence for dataclasses.
        This brings specific assumptions
        - `self` is the native inmanta type
        - `type` is translated form the python domain
        - 'Any' means: trust me, it will be fine (i.e. always corresponds)
        """
        raise NotImplementedError()

    def as_python_type_string(self) -> "str | None":
        """
        Return a python type that can capture the values of this inmanta type
        As a string

        Returns None if this is not possible.

        Used only on the plugin boundary
        """

    def has_custom_to_python(self) -> bool:
        """
        Indicates if a special to_python conversion should be used

        Used only on the plugin boundary
        """
        return False

    def to_python(self, instance: object, *, path: str) -> object:
        """
        Convert an instance of this type to its python form

        should only be called if has_custom_to_python is True
        the instance must be valid according to the validate method

        :param path: The path, relative to the plugin's namespace, where this object occurs.

        :raises RuntimeException: If there is a translation error.
        """
        return instance

    def issupertype(self, other: "Type") -> bool:
        """
        Returns True iff this DSL type is a supertype of the other type in a non-trivial way.
        Raises NotImplementedError if this DSL type has no special supertyping rules (i.e. most types).
        """
        raise NotImplementedError()

    def issubtype(self, other: "Type") -> bool:
        """
        Returns True iff this DSL type is a subtype of the other type.
        False negatives may be unavoidable in complex cases. Does not return false positives.
        """
        if self == other:
            return True
        try:
            return other.issupertype(self)
        except NotImplementedError:
            return False

    def __eq__(self, other: object) -> bool:
        if type(self) != Type:  # noqa: E721
            # Not for children
            return NotImplemented
        return type(self) == type(other)  # noqa: E721

    def __hash__(self) -> int:
        return hash(type(self))


class ReferenceType(Type):
    """
    The type of a reference to something of type element_type

    e.g ReferenceType(Integer()) represents a reference to an int
    """

    def __init__(self, element_type: Type) -> None:
        """
        :param element_type: the type we refer to
        """
        super().__init__()
        assert not element_type.supports_references()
        self.element_type = element_type
        self.is_dataclass = False
        if element_type.is_entity():
            # Can not be typed more strictly due to import loops
            # The root cause of the problem is the to_dsl method, which is required by entity and plugin
            # these are types, but to_dsl also constructs types.
            # i.e. we can't layer the type, entity and plugin domain any more
            if element_type.get_paired_dataclass() is None:
                raise TypingException(
                    None,
                    f"References to entities must always be references to dataclasses."
                    f" Got {element_type}, which is not a dataclass",
                )

            self.is_dataclass = True

    def validate(self, value: Optional[object]) -> bool:
        ref: Optional[references.Reference[references.RefValue]] = references.unwrap_reference(value)
        if ref is not None:
            assert ref._model_type is not None
            if ref._model_type.issubtype(self.element_type):
                return True

        raise TypingException(None, f"Invalid value: {value} is not a subtype of {self}")

    def has_custom_to_python(self) -> bool:
        return self.is_dataclass

    def to_python(self, instance: object, *, path: str) -> object:
        result: Optional[references.Reference[references.RefValue]] = references.unwrap_reference(instance)
        # wouldn't have passed validate otherwise
        assert result is not None
        return result

    def type_string_internal(self) -> str:
        return f"Reference[{self.element_type.type_string_internal()}]"

    def is_attribute_type(self) -> bool:
        return self.element_type.is_attribute_type()

    def get_base_type(self) -> "Type":
        return self.element_type

    def get_no_reference(self) -> "Type":
        return self.element_type

    def corresponds_to(self, type: "Type") -> bool:
        if builtins.type(type) != builtins.type(self):
            return False

        return self.element_type.corresponds_to(type.element_type)

    def as_python_type_string(self) -> "str | None":
        return f"Reference[{self.element_type.as_python_type_string()}]"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ReferenceType):
            return False
        return other.element_type == self.element_type

    def __hash__(self) -> int:
        return hash((type(self), self.element_type))

    def issupertype(self, other: "Type") -> bool:
        if not isinstance(other, ReferenceType):
            return False
        return self.element_type.issupertype(other.element_type)


class OrReferenceType(Type):
    """
    This class represents the shorthand for Reference[T] | T

    It produces a cleaner output for exceptions

    create_unions will compact unions to use this class when relevant
    """

    def __init__(self, element_type: Type) -> None:
        self.element_type: Type = element_type
        self.reference_type: ReferenceType = ReferenceType(element_type)

    def validate(self, value: Optional[object]) -> bool:
        # We validate that the value is either a reference of the base type or the base type
        if references.unwrap_reference(value) is not None:
            # Validate that we are the reference
            return self.reference_type.validate(value)
        else:
            # Validate that we are the base type
            return self.element_type.validate(value)

    def has_custom_to_python(self) -> bool:
        return self.reference_type.has_custom_to_python() or self.element_type.has_custom_to_python()

    def to_python(self, instance: object, *, path: str) -> object:
        try:
            self.reference_type.validate(instance)
        except RuntimeException:
            if self.element_type.has_custom_to_python():
                return self.element_type.to_python(instance, path=path)
            else:
                return DynamicProxy.return_value(instance, context=ProxyContext(path=path))
        else:
            return self.reference_type.to_python(instance, path=path)

    def type_string(self) -> Optional[str]:
        # unfortunately, this type is used (by necessity) for attribute types.
        return self.element_type.type_string()

    def type_string_internal(self) -> str:
        element = self.element_type.type_string_internal()
        return f"Reference[{element}] | {element}"

    def as_python_type_string(self) -> "str | None":
        # Can't be expressed in the model
        return f"Reference[{self.element_type.as_python_type_string()}] | {self.element_type.as_python_type_string()}"

    def __eq__(self, other):
        if not isinstance(other, OrReferenceType):
            return False
        return other.element_type == self.element_type

    def __hash__(self) -> int:
        return hash((type(self), self.element_type))

    def is_attribute_type(self) -> bool:
        return self.element_type.is_attribute_type()

    def get_base_type(self) -> "Type":
        return self.element_type

    def get_no_reference(self) -> "Type":
        return self.element_type

    def corresponds_to(self, type: "Type") -> bool:
        # The model always allow reference, we allow the type in the python domain to be tighter
        if isinstance(type, Any):
            return True
        if self.element_type.corresponds_to(type):
            return True
        if isinstance(type, OrReferenceType):
            return self.element_type.corresponds_to(type.element_type)
        return False

    def issupertype(self, other: "Type") -> bool:
        if not isinstance(other, (ReferenceType, OrReferenceType)):
            return self.element_type.issupertype(other)
        return self.element_type.issupertype(other.element_type)


class NamedType(Type, Named):
    def get_double_defined_exception(self, other: "NamedType") -> "DuplicateException":
        """produce an error message for this type"""
        raise DuplicateException(self, other, "Type %s is already defined" % (self.get_full_name()))

    def type_string(self) -> str:
        return self.get_full_name()

    def type_string_internal(self) -> str:
        return self.type_string()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NamedType):
            return False
        return self.get_full_name() == other.get_full_name()

    def __hash__(self) -> int:
        return hash((type(self), self.get_full_name()))


class Null(Type):
    """
    This custom type is used for the validation of plugins which only
    accept null as an argument or return value.
    """

    def validate(self, value: Optional[object]) -> bool:
        if isinstance(value, NoneValue):
            return True

        raise RuntimeException(None, f"Invalid value '{value}', expected {self.type_string()}")

    def type_string(self) -> str:
        return "null"

    def type_string_internal(self) -> str:
        return self.type_string()

    def as_python_type_string(self) -> "str | None":
        return "None"

    def corresponds_to(self, type: Type) -> bool:
        return isinstance(type, (Null, Any))

    def has_custom_to_python(self) -> bool:
        return False

    def __eq__(self, other: object) -> bool:
        return type(self) == type(other)  # noqa: E721

    def get_location(self) -> Optional[Location]:
        return None

    def __hash__(self) -> int:
        return hash(self.type_string())


@stable_api
class NullableType(Type):
    """
    Represents a nullable type in the Inmanta :term:`DSL`. For example ``NullableType(Number())`` represents ``number?``.
    """

    def __init__(self, element_type: Type) -> None:
        Type.__init__(self)
        self.element_type: Type = element_type

    def validate(self, value: Optional[object]) -> bool:
        if isinstance(value, NoneValue):
            return True

        return self.element_type.validate(value)

    def _wrap_type_string(self, string: str) -> str:
        return f"({string})?" if " " in string else f"{string}?"

    def type_string(self) -> Optional[str]:
        base_type_string: Optional[str] = self.element_type.type_string()
        return None if base_type_string is None else self._wrap_type_string(base_type_string)

    def type_string_internal(self) -> str:
        return self._wrap_type_string(self.element_type.type_string_internal())

    def normalize(self) -> None:
        self.element_type.normalize()

    def get_base_type(self) -> Type:
        return self.element_type.get_base_type()

    def get_no_reference(self) -> "Type":
        base = self.element_type.get_no_reference()
        if base is self.element_type:
            return self
        return NullableType(base)

    def with_base_type(self, base_type: Type) -> Type:
        return NullableType(self.element_type.with_base_type(base_type))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NullableType):
            return NotImplemented
        return self.element_type == other.element_type

    def __hash__(self) -> int:
        return hash((type(self), self.element_type))

    def is_attribute_type(self) -> bool:
        return self.element_type.is_attribute_type()

    def corresponds_to(self, type: Type) -> bool:
        if isinstance(type, Any):
            return True
        if not isinstance(type, NullableType):
            return False
        return self.element_type.corresponds_to(type.element_type)

    def as_python_type_string(self) -> "str | None":
        return f"{self.element_type.as_python_type_string()} | None"

    def to_python(self, instance: object, *, path: str) -> object:
        if isinstance(instance, NoneValue):
            return None
        return self.element_type.to_python(instance, path=path)

    def has_custom_to_python(self) -> bool:
        return self.element_type.has_custom_to_python()

    def issupertype(self, other: "Type") -> bool:
        return isinstance(other, Null) or other.issubtype(self.element_type)


class Any(Type):
    """
    this type represents the any type, similar to Python's

    the Any class itself is neither the top nor the bottom type in the type hierarchy.
    """

    def validate(self, value: Optional[object]) -> bool:
        try:
            return super().validate(value)
        except UnexpectedReference as e:
            raise UnexpectedReference(
                reference=e.reference,
                # custom error message for the "object" / "any" type
                message=(
                    "References are not allowed for values of type `object`. While the `object` Python type is technically"
                    " compatible with any value, references are considered special DSL values and are therefore guarded so"
                    " that they don't accidentally show up where they are not expected. To work with references,"
                    " explicitly declare support with a `object | Reference` annotation."
                ),
            )

    def corresponds_to(self, type: Type) -> bool:
        return True

    def as_python_type_string(self) -> "str | None":
        return "object"

    def has_custom_to_python(self) -> bool:
        return False

    def type_string_internal(self) -> str:
        return "any"

    def __eq__(self, other: object) -> bool:
        return type(self) == type(other)  # noqa: E721

    def issupertype(self, other: "Type") -> bool:
        if other == self:
            return False
        return True

    def __hash__(self):
        # Could be any unique value
        return 3141156432848106867


def cast_not_implemented(value: Optional[object]) -> object:
    raise NotImplementedError("Can not cast to Primitive")


@stable_api
class Primitive(Type):
    """
    Abstract base class representing primitive types.
    """

    def __init__(self) -> None:
        Type.__init__(self)
        self.cast_function: Callable[[Optional[object]], object] = cast_not_implemented

    def cast(self, value: Optional[object]) -> object:
        """
        Cast a value to this type. If the value can not be cast, raises a :py:class:`inmanta.ast.RuntimeException`.
        """
        if isinstance(value, Unknown):
            # propagate unknowns
            return value

        try:
            return self.cast_function(value)
        except (ValueError, TypeError):
            raise RuntimeException(None, f"Failed to cast '{value}' to {self}")

    def type_string_internal(self) -> str:
        return "Primitive"

    def corresponds_to(self, type: "Type") -> bool:
        if isinstance(type, Any):
            return True
        return type == self

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other)

    def __hash__(self) -> int:
        return hash(self.type_string())

    def has_custom_to_python(self) -> bool:
        # All primitives can be trivially converted
        return False

    def is_attribute_type(self) -> bool:
        return True

    def get_location(self) -> Optional[Location]:
        # Override to skip the null check in the parent class
        return None

    def issupertype(self, other: "Type") -> bool:
        return False


@stable_api
class Number(Primitive):
    """
    This class represents an integer or a float in the configuration model.
    """

    def __init__(self) -> None:
        Primitive.__init__(self)
        self.cast_function = float

    def cast(self, value: Optional[object]) -> object:
        """
        Attempts to cast a given value to an int or a float.
        """
        # Keep precision: cast to an int only if it already is an int
        if isinstance(value, int):
            return int(value)
        return super().cast(value)

    def validate(self, value: Optional[object]) -> bool:
        """
        Validate the given value to check if it satisfies the constraints
        associated with this type
        """
        super().validate(value)
        if isinstance(value, AnyType):
            return True

        if not isinstance(value, numbers.Number):
            raise RuntimeException(None, f"Invalid value '{value}', expected {self.type_string()}")

        return True  # allow this function to be called from a lambda function

    def get_location(self) -> None:
        return None

    def type_string(self) -> str:
        return "number"

    def type_string_internal(self) -> str:
        return self.type_string()

    def as_python_type_string(self) -> "str | None":
        return "numbers.Number"

    def corresponds_to(self, type: "Type") -> bool:
        return isinstance(type, (Any, Float, Integer, Number))

    def issupertype(self, other: "Type") -> bool:
        return isinstance(other, (Float, Integer))


@stable_api
class Float(Primitive):
    """
    This class is an alias for the Number class and represents a float in
    the configuration model.
    """

    def __init__(self) -> None:
        Primitive.__init__(self)
        self.cast_function = float

    def validate(self, value: Optional[object]) -> bool:
        """
        Validate the given value to check if it satisfies the constraints
        associated with this type
        """
        super().validate(value)
        if isinstance(value, AnyType):
            return True

        if not isinstance(value, float):
            raise RuntimeException(None, f"Invalid value '{value}', expected {self.type_string()}")
        return True  # allow this function to be called from a lambda function

    def get_location(self) -> None:
        return None

    def type_string(self) -> str:
        return "float"

    def type_string_internal(self) -> str:
        return self.type_string()

    def as_python_type_string(self) -> "str | None":
        return "float"


@stable_api
class Integer(Primitive):
    """
    An instance of this class represents the int type in the configuration model.
    """

    def __init__(self) -> None:
        super().__init__()
        self.cast_function = int

    def validate(self, value: Optional[object]) -> bool:
        """
        Validate the given value to check if it satisfies the constraints
        associated with this type
        """
        super().validate(value)
        if isinstance(value, AnyType):
            return True

        if not isinstance(value, numbers.Integral):
            raise RuntimeException(None, f"Invalid value '{value}', expected {self.type_string()}")
        return True  # allow this function to be called from a lambda function

    def type_string(self) -> str:
        return "int"

    def type_string_internal(self) -> str:
        return "int"

    def as_python_type_string(self) -> "str | None":
        return "int"


@stable_api
class Bool(Primitive):
    """
    This class represents a simple boolean that can hold true or false.
    """

    def __init__(self) -> None:
        Primitive.__init__(self)
        self.cast_function = bool

    def validate(self, value: Optional[object]) -> bool:
        """
        Validate the given value to check if it satisfies the constraints
        associated with this type
        """
        super().validate(value)
        if isinstance(value, AnyType):
            return True
        if isinstance(value, bool):
            return True
        raise RuntimeException(None, f"Invalid value '{value}', expected {self.type_string()}")

    def cast(self, value: Optional[object]) -> object:
        # this is a bit odd, in that is accepts None, but it has always been so
        return super().cast(value if not isinstance(value, NoneValue) else None)

    def type_string(self) -> str:
        return "bool"

    def type_string_internal(self) -> str:
        return self.type_string()

    def get_location(self) -> None:
        return None

    def as_python_type_string(self) -> "str | None":
        return "bool"


@stable_api
class String(Primitive):
    """
    This class represents a string type in the configuration model.
    """

    def __init__(self) -> None:
        Primitive.__init__(self)
        self.cast_function = str

    def validate(self, value: Optional[object]) -> bool:
        """
        Validate the given value to check if it satisfies the constraints
        associated with this type
        """
        super().validate(value)
        if isinstance(value, AnyType):
            return True
        if not isinstance(value, str):
            raise RuntimeException(None, f"Invalid value '{value}', expected {self.type_string()}")

        return True

    def cast(self, value: Optional[object]) -> object:
        if value is True:
            return "true"
        if value is False:
            return "false"
        return super().cast(value)

    def type_string(self) -> str:
        return "string"

    def type_string_internal(self) -> str:
        return self.type_string()

    def get_location(self) -> None:
        return None

    def as_python_type_string(self) -> "str | None":
        return "str"


@stable_api
class List(Type):
    """
    Instances of this class represent a list type containing any types of values.
    This class refers to the list type used in plugin annotations. For the list type in the Inmanta DSL, see `LiteralList`.
    """

    def __init__(self) -> None:
        Type.__init__(self)

    def validate(self, value: Optional[object]) -> bool:
        if value is None:
            return True

        if isinstance(value, AnyType):
            return True

        if not isinstance(value, list):
            raise TypingException(None, f"Invalid value '{value}', expected {self.type_string()}")

        return True

    def type_string(self) -> str:
        # This is not a type in the model, but it is used in plugin annotations, which are also part of the DSL.
        return "list"

    def type_string_internal(self) -> str:
        return "list"

    def get_location(self) -> None | Location:
        return None

    def corresponds_to(self, type: Type) -> bool:
        # Unreachable, the model can't specify this
        if isinstance(type, Any):
            return True

        return isinstance(type, List)

    def as_python_type_string(self) -> "str | None":
        return "list[object]"

    def __eq__(self, other: object) -> bool:
        return type(self) == type(other)  # noqa: E721

    def __hash__(self) -> int:
        return hash(type(self))

    def issubtype(self, other: "Type") -> bool:
        if isinstance(other, Any):
            return True
        return isinstance(other, List)

    def issupertype(self, other: "Type") -> bool:
        raise NotImplementedError()


@stable_api
class TypedList(List):
    """
    Instances of this class represent a list type containing any values of type element_type.
    For example `TypedList(Number())` represents `number[]`.
    """

    def __init__(self, element_type: Type) -> None:
        List.__init__(self)
        self.element_type: Type = element_type

    def normalize(self) -> None:
        self.element_type.normalize()

    def validate(self, value: Optional[object]) -> bool:
        List.validate(self, value)

        assert isinstance(value, list)
        for element in value:
            self.element_type.validate(element)

        return True

    def get_no_reference(self) -> "Type":
        base = self.element_type.get_no_reference()
        if base is self.element_type:
            return self
        return TypedList(base)

    def _wrap_type_string(self, string: str) -> str:
        return f"({string})[]" if " " in string else f"{string}[]"

    def type_string(self) -> Optional[str]:
        element_type_string = self.element_type.type_string()
        return None if element_type_string is None else self._wrap_type_string(element_type_string)

    def type_string_internal(self) -> str:
        return self._wrap_type_string(self.element_type.type_string_internal())

    def get_location(self) -> Location | None:
        return self.element_type.get_location()

    def get_base_type(self) -> Type:
        return self.element_type

    def with_base_type(self, base_type: Type) -> Type:
        return TypedList(base_type)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TypedList):
            return NotImplemented
        return self.element_type == other.element_type

    def __hash__(self) -> int:
        return hash((type(self), self.element_type))

    def is_attribute_type(self) -> bool:
        return self.element_type.is_attribute_type()

    def corresponds_to(self, type: Type) -> bool:
        if isinstance(type, Any):
            return True

        if not isinstance(type, TypedList):
            # The other list is untyped, so we are not equivalent
            return False

        return self.element_type.corresponds_to(type.element_type)

    def as_python_type_string(self) -> "str | None":
        return f"list[{self.element_type.as_python_type_string()}]"

    def to_python(self, instance: object, *, path: str) -> object:
        if not isinstance(instance, Sequence):
            # should not happen, pre-condition
            raise TypeError(f"This method can only be called on iterables, not on {type(instance)}")
        return [self.element_type.to_python(element, path=f"{path}[{i}]") for i, element in enumerate(instance)]

    def has_custom_to_python(self) -> bool:
        return self.element_type.has_custom_to_python()

    def issubtype(self, other: "Type") -> bool:
        if isinstance(other, Any):
            return True
        if not isinstance(other, TypedList):
            return False
        return self.element_type.issubtype(other.element_type)

    def issupertype(self, other: "Type") -> bool:
        if not isinstance(other, TypedList):
            return False
        return self.element_type.issupertype(other.element_type)


@stable_api
class LiteralList(TypedList):
    """
    Instances of this class represent a list type containing only :py:class:`Literal` values.
    This is the `list` type in the :term:`DSL`
    """

    def __init__(self) -> None:
        TypedList.__init__(self, Literal())

    def type_string(self) -> str:
        return "list"

    def get_base_type(self) -> Type:
        # The `list` type is not multi, thus it is the base type itself
        return self

    def with_base_type(self, base_type: Type) -> Type:
        return self

    def has_custom_to_python(self) -> bool:
        return False

    def get_no_reference(self) -> Type:
        return self

    def is_attribute_type(self) -> bool:
        return True

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LiteralList):
            return NotImplemented
        return True


@stable_api
class Dict(Type):
    """
    Instances of this class represent a dict type with any types of values.
    """

    def __init__(self) -> None:
        Type.__init__(self)

    def validate(self, value: Optional[object]) -> bool:
        """
        Validate the given value to check if it satisfies the constraints
        associated with this type
        """
        if isinstance(value, AnyType):
            return True

        if value is None:
            return True

        if not isinstance(value, dict):
            raise RuntimeException(None, f"Invalid value '{value}', expected {self.type_string()}")

        return True

    def type_string_internal(self) -> str:
        return "Dict"

    def type_string(self) -> str:
        return "dict"

    def get_location(self) -> None | Location:
        return None

    def as_python_type_string(self) -> "str | None":
        return "dict"

    def corresponds_to(self, type: Type) -> bool:
        if isinstance(type, Any):
            return True
        return isinstance(type, Dict)

    def issubtype(self, other: "Type") -> bool:
        if isinstance(other, Any):
            return True
        return isinstance(other, Dict)

    def issupertype(self, other: "Type") -> bool:
        raise NotImplementedError()

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other)

    def __hash__(self) -> int:
        return hash(type(self))


@stable_api
class TypedDict(Dict):
    """
    Instances of this class represent a dict type containing only values of type element_type.
    """

    def __init__(self, element_type: Type) -> None:
        Dict.__init__(self)
        self.element_type: Type = element_type

    def normalize(self) -> None:
        self.element_type.normalize()

    def validate(self, value: Optional[object]) -> bool:
        Dict.validate(self, value)

        assert isinstance(value, dict)
        for element in value.values():
            self.element_type.validate(element)

        return True

    def type_string_internal(self) -> str:
        return "dict[string, %s]" % self.element_type.type_string_internal()

    def get_location(self) -> Location | None:
        return self.element_type.get_location()

    def corresponds_to(self, type: Type) -> bool:
        if isinstance(type, Any):
            return True

        if not isinstance(type, Dict):
            return False

        if not isinstance(type, TypedDict):
            # Untyped dict is fine
            return True

        return self.element_type.corresponds_to(type.element_type)

    def as_python_type_string(self) -> "str | None":
        return f"dict[str, {self.element_type.as_python_type_string()}]"

    def has_custom_to_python(self) -> bool:
        return self.element_type.has_custom_to_python()

    def to_python(self, instance: object, *, path: str) -> object:
        assert isinstance(instance, dict)
        base = self.element_type
        return {k: base.to_python(v, path=f"{path}[{k}]") for k, v in instance.items()}

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TypedDict):
            return NotImplemented
        return self.element_type == other.element_type

    def __hash__(self) -> int:
        return hash((type(self), self.element_type))

    def issubtype(self, other: "Type") -> bool:
        if isinstance(other, Any):
            return True
        if not isinstance(other, TypedDict):
            return False
        return self.element_type.issubtype(other.element_type)

    def issupertype(self, other: "Type") -> bool:
        if not isinstance(other, TypedDict):
            return False
        return self.element_type.issupertype(other.element_type)

    def get_no_reference(self) -> "Type":
        base = self.element_type.get_no_reference()
        if base is self.element_type:
            return self
        return TypedDict(base)


@stable_api
class LiteralDict(TypedDict):
    """
    Instances of this class represent a dict type containing only :py:class:`Literal` values.
    This is the `dict` type in the :term:`DSL`
    """

    def __init__(self) -> None:
        TypedDict.__init__(self, Literal())

    def type_string(self) -> str:
        return "dict"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LiteralDict):
            return NotImplemented
        return True

    def has_custom_to_python(self) -> bool:
        return False

    def get_no_reference(self) -> Type:
        return self

    def is_attribute_type(self) -> bool:
        return True


@dataclasses.dataclass
class BaseOrRef:
    """Small helper to sort types and their associated reference"""

    has_base: Type | None = None
    has_ref: ReferenceType | None = None

    def convert(self) -> Type:
        if self.has_ref is not None and self.has_base is not None:
            return OrReferenceType(self.has_base)
        if self.has_ref is not None:
            return self.has_ref
        assert self.has_base is not None
        return self.has_base


def create_union(types: Sequence[Type]) -> Type:
    """
    Normalize the union:
     - Nullable is the outer type if applicable
     - Nested unions are flattened
     - Reference[T] | T becomes OrReference[T] (for cleaner output)
     - Single item union return just the item
    """
    worklist = deque(types)
    sorted_types: dict[Type, BaseOrRef] = defaultdict(BaseOrRef)
    nullable = False
    seen: set[Type] = set()
    while worklist:
        current = worklist.popleft()
        if current in seen:
            continue
        seen.add(current)
        match current:
            case Union():
                worklist.extend(current.types)
            case NullableType():
                nullable = True
                worklist.append(current.element_type)
            case Null():
                nullable = True
            case ReferenceType():
                sorted_types[current.element_type].has_ref = current
            case _:
                sorted_types[current].has_base = current

    bases = [bor.convert() for bor in sorted_types.values()]
    if len(bases) == 1:
        base_union: Type = bases[0]
    else:
        base_union = Union(bases)

    if nullable:
        return NullableType(base_union)
    else:
        return base_union


@stable_api
class Union(Type):
    """
    Instances of this class represent a union of multiple types.
    """

    def __init__(self, types: Sequence[Type]) -> None:
        Type.__init__(self)
        self.types: Sequence[Type] = types

    def get_base_type(self) -> "Type":
        return self

    def get_no_reference(self) -> "Type":
        # CACHE!!!
        bases = {type.get_no_reference() for type in self.types}
        if len(bases) == 1:
            return next(iter(bases))
        if bases == set(self.types):
            return self
        return Union(list(bases))

    def validate(self, value: object) -> bool:
        for typ in self.types:
            try:
                if typ.validate(value):
                    return True
            except RuntimeException:
                pass
        raise TypingException(None, f"Invalid value '{value}', expected {self}")

    def type_string_internal(self) -> str:
        return "Union[%s]" % ",".join(t.type_string_internal() for t in self.types)

    def as_python_type_string(self) -> "str | None":
        types = [tp.as_python_type_string() for tp in self.types]

        effective_types = [tp for tp in types if tp is not None]
        if len(types) != len(effective_types):
            # One is not converted
            return None
        return " | ".join(effective_types)

    def has_custom_to_python(self) -> bool:
        return any(tp.has_custom_to_python() for tp in self.types)

    def to_python(self, instance: object, *, path: str) -> object:
        """
        Construct a python object for this instance
        """
        # At this point, we have two pre-conditions
        # 1. instance is an instance of one of our types
        # 2. one of our types has_custom_to_python set
        for tp in self.types:
            # For each of out types
            try:
                # Find if this instance is of that type
                if tp.validate(instance):
                    # It is of this type, does it require custom conversion?
                    if tp.has_custom_to_python():
                        # Custom conversion
                        return tp.to_python(instance, path=path)
                    else:
                        # Default conversion
                        return DynamicProxy.return_value(instance, context=ProxyContext(path=path))
            except (UnsetException, MultiUnsetException):
                # let these exceptions with special meaning through
                raise
            except RuntimeException:
                # Validate fails, up to the next one
                pass
        # Due to the invariants, this can't happen
        # One of the types HAS to match validate
        assert False

    def corresponds_to(self, type: Type) -> bool:
        if isinstance(type, Union):
            types = list(type.types)
        else:
            types = [type]

        unmatched = list(types)
        for type in self.types:
            for other_type in types:
                if type.corresponds_to(other_type):
                    unmatched.remove(other_type)
                    break
            else:
                # type did not match anything
                return False
        return not unmatched

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Union):
            return NotImplemented
        return self.types == other.types

    def is_attribute_type(self) -> bool:
        # It can not strictly speaking be used as an attribute type
        # But, if all member are is_attribute_type this is equivalent to either Nullable or Literal
        return all(tp.is_attribute_type() for tp in self.types)

    def get_location(self) -> Optional[Location]:
        # We don't know what location to use...
        return None

    def issupertype(self, other: "Type") -> bool:
        return any(other.issubtype(tp) for tp in self.types)

    def issubtype(self, other: "Type") -> bool:
        return all(element_type.issubtype(other) for element_type in self.types)

    def __hash__(self) -> int:
        return hash((3141156432848106868, *self.types))


@stable_api
class Literal(Union):
    """
    Instances of this class represent a literal in the configuration model. A literal is a primitive or a list or dict
    where all values are literals themselves.
    """

    def __init__(self) -> None:
        Union.__init__(
            self, [NullableType(Float()), Number(), Bool(), String(), TypedList(self), TypedDict(self), ReferenceType(self)]
        )

    def type_string_internal(self) -> str:
        return "Literal"

    def as_python_type_string(self) -> "str | None":
        # Keep it simple
        return "object"

    def is_attribute_type(self) -> bool:
        return True

    def corresponds_to(self, type: Type) -> bool:
        if isinstance(type, Any):
            return True

        # Infinite recursive type, avoid the mess
        # We allow any primitive
        return type.is_attribute_type()

    def issupertype(self, other: "Type") -> bool:
        return other != self and other.is_attribute_type()

    def issubtype(self, other: "Type") -> bool:
        return Type.issubtype(self, other)

    def has_custom_to_python(self) -> bool:
        return False

    def get_no_reference(self) -> "Type":
        return self


@stable_api
class ConstraintType(NamedType):
    """
    A type that is based on a primitive type but defines additional constraints on this type.
    These constraints only apply on the value of the type.
    """

    def __init__(self, namespace: Namespace, name: str) -> None:
        NamedType.__init__(self)

        # It is easy to assume that get_base_type return self.basetype
        # It doesn't and shouldn't
        # This field would better be called element_type, but that would break backward compatibility
        # Is it also assumed to NEVER be a reference type
        self.basetype: Optional[Type] = None  # : ConstrainableType
        self._constraint: Callable[[object], object] | None = None
        self.name: str = name
        self.namespace: Namespace = namespace
        self.comment: Optional[str] = None
        self.expression: Optional["ExpressionStatement"] = None

    def normalize(self) -> None:
        assert self.expression is not None
        self.expression.normalize()

    def set_constraint(self, expression: "ExpressionStatement") -> None:
        """
        Set the constraint for this type. This baseclass for constraint
        types requires the constraint to be set as a regex that can be
        compiled.
        """
        self.expression = expression
        self._constraint = create_function(self, expression)

    def get_constraint(self) -> "ExpressionStatement | None":
        """
        Get the string representation of the constraint
        """
        return self.expression

    constraint = property(get_constraint, set_constraint)

    def validate(self, value: Optional[object]) -> bool:
        """
        Validate the given value to check if it satisfies the constraint and
        the basetype.
        """
        if isinstance(value, AnyType):
            return True

        assert self.basetype is not None
        self.basetype.validate(value)

        assert self._constraint is not None
        assert self.expression is not None
        if not self._constraint(value):
            raise RuntimeException(
                self, f"Invalid value {repr(value)}, does not match constraint `{self.expression.pretty_print()}`"
            )

        return True

    def type_string(self) -> str:
        return f"{self.namespace}::{self.name}"

    def type_string_internal(self) -> str:
        return self.type_string()

    def get_full_name(self) -> str:
        return self.namespace.get_full_name() + "::" + self.name

    def get_namespace(self) -> "Namespace":
        return self.namespace

    def get_double_defined_exception(self, other: "NamedType") -> DuplicateException:
        return DuplicateException(self, other, "TypeConstraint %s is already defined" % (self.get_full_name()))

    def has_custom_to_python(self) -> bool:
        # Substitute for base type for now
        assert self.basetype is not None
        return self.basetype.has_custom_to_python()

    def corresponds_to(self, type: Type) -> bool:
        if self is type:
            # To avoid comparing the expression, we evaluate on exact equality, as typedefs are uniquely defined
            return True
        if isinstance(type, Any):
            return True
        assert self.basetype is not None
        # Same basetype is sufficiently close
        return self.basetype.corresponds_to(type)

    def as_python_type_string(self) -> "str | None":
        assert self.basetype is not None
        return self.basetype.as_python_type_string()

    def to_python(self, instance: object, *, path: str) -> object:
        assert self.basetype is not None
        return self.basetype.to_python(instance, path=path)

    def issubtype(self, other: "Type") -> bool:
        assert self.basetype is not None
        return super().issubtype(other) or self.basetype.issubtype(other)


def create_function(tp: ConstraintType, expression: "ExpressionStatement") -> Callable[[object], object]:
    """
    Function that returns a function that evaluates the given expression.
    The generated function accepts the unbound variables in the expression
    as arguments.
    """

    def function(*args, **kwargs):
        """
        A function that evaluates the expression
        """
        if len(args) != 1:
            raise NotImplementedError()

        try:
            return expression.execute_direct({"self": args[0]})
        except NotFoundException as e:
            e.msg = "Unable to resolve `%s`: a type constraint can not reference variables." % e.stmt.name
            raise e

    return function


TYPES: dict[str, Type] = {  # Part of the stable API
    "string": String(),
    "float": Float(),
    "number": Number(),
    "int": Integer(),
    "bool": Bool(),
    "list": LiteralList(),
    "dict": LiteralDict(),
}
"""
    Maps Inmanta :term:`DSL` types to their internal representation. For each key, value pair, `value.type_string()` is
    guaranteed to return key.
"""


@stable_api
def resolve_type(locatable_type: LocatableString, resolver: Namespace) -> Type:
    """
    Convert a locatable type string, into a real inmanta type, that can be used for validation.

    :param locatable_type: An object pointing to the type expression.
    :param resolver: The namespace that can be used to resolve the type expression
    """
    # quickfix issue #1774
    allowed_element_type: Type = Any()
    if locatable_type.value == "list":
        return List()
    if locatable_type.value == "dict":
        return TypedDict(allowed_element_type)

    # stack of transformations to be applied to the base inmanta_type.Type
    # transformations will be applied right to left
    transformation_stack: List[Callable[[Type], Type]] = []

    if locatable_type.value.endswith("?"):
        # We don't want to modify the object we received as argument
        locatable_type = copy.copy(locatable_type)
        locatable_type.value = locatable_type.value[0:-1]
        transformation_stack.append(NullableType)

    if locatable_type.value.endswith("[]"):
        # We don't want to modify the object we received as argument
        locatable_type = copy.copy(locatable_type)
        locatable_type.value = locatable_type.value[0:-2]
        transformation_stack.append(TypedList)

    return functools.reduce(
        lambda acc, transform: transform(acc), reversed(transformation_stack), resolver.get_type(locatable_type)
    )



"""
Copyright 2022 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    schema = f"""
    ALTER TABLE ONLY public.resource
        DROP CONSTRAINT resource_pkey;

    ALTER TABLE ONLY public.resource
        ADD PRIMARY KEY (environment, model, resource_id);

    ALTER TABLE public.resource
        DROP COLUMN resource_version_id;

    DROP INDEX resourceaction_resource_version_ids_index;

    CREATE TYPE resource_id_version_pair AS (resource_id character varying, version int);

    UPDATE public.resource SET
        provides = {query_part_convert_provides('provides')},
        attributes = {query_part_convert_requires('attributes')};

    """

    # Deep rewrite of the data
    # convert resource_version_id into resource_id
    await connection.execute(schema)


def query_part_convert_rvid_to_rid(from_value: str) -> str:
    """
    Query part to actually convert a resource version id to a resource id

    split off for easier testing
    """
    return f"REGEXP_REPLACE({from_value}, ',v=[0-9]+$', '')"


def query_part_convert_requires(from_value: str) -> str:
    """
    Query part to convert requires, which is in a jsonb dict

    split off for easier testing
    """
    # get the value out of the dict
    project = f"{from_value}->'requires'"

    # make conversion on one element
    convert = query_part_convert_rvid_to_rid("element")

    # loop over the array and convert
    collect = f"""
     (select
         jsonb_agg(
            {convert}
         )
     from jsonb_array_elements_text({project}) element)"""

    # set the value and handle the case where the requires key doesn't exist
    return f"coalesce(jsonb_set({from_value}, '{{requires}}', {collect}), {from_value})"


def query_part_convert_provides(from_value: str) -> str:
    """
    Query part to convert provides, which is a varchar array

    split off for easier testing
    """
    # make conversion on one element
    convert = query_part_convert_rvid_to_rid("element")

    # loop over the array and convert
    collect = f"""
        coalesce(
            (select
                 array_agg(
                   {convert}
                )
            from unnest({from_value}) element),
            ARRAY[]::character varying[]
        )
        """

    return collect
